# Copyright 2021 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=no-member, import-outside-toplevel

"""
This module defines gates and operations that can be applied to quantum modes to construct a quantum circuit.
"""

from typing import List, Optional, Sequence, Tuple, Union
import numpy as np

from mrmustard import settings
from mrmustard.physics import gaussian, fock
from mrmustard.utils.typing import ComplexMatrix, RealMatrix
from mrmustard import math
from mrmustard.math.parameters import update_orthogonal, update_symplectic, update_unitary
from .abstract import Channel, Unitary, State
from .utils import make_parameter

__all__ = [
    "Dgate",
    "Sgate",
    "Rgate",
    "Pgate",
    "Ggate",
    "BSgate",
    "MZgate",
    "S2gate",
    "CZgate",
    "CXgate",
    "Interferometer",
    "RealInterferometer",
    "Attenuator",
    "Amplifier",
    "AdditiveNoise",
    "PhaseNoise",
]


class Dgate(Unitary):
    r"""Displacement gate.

    If ``len(modes) > 1`` the gate is applied in parallel to all of the modes provided.

    If a parameter is a single float, the parallel instances of the gate share that parameter.

    To apply mode-specific values use a list of floats. One can optionally set bounds for each
    parameter, which the optimizer will respect.

    Args:
        x (float or List[float]): the list of displacements along the x axis
        x_bounds (float, float): bounds for the displacement along the x axis
        x_trainable (bool): whether x is a trainable variable
        y (float or List[float]): the list of displacements along the y axis
        y_bounds (float, float): bounds for the displacement along the y axis
        y_trainable bool: whether y is a trainable variable
        modes (optional, List[int]): the list of modes this gate is applied to
    """

    is_gaussian = True
    short_name = "D"
    parallelizable = True

    def __init__(
        self,
        x: Union[float, List[float]] = 0.0,
        y: Union[float, List[float]] = 0.0,
        x_trainable: bool = False,
        y_trainable: bool = False,
        x_bounds: Tuple[Optional[float], Optional[float]] = (None, None),
        y_bounds: Tuple[Optional[float], Optional[float]] = (None, None),
        modes: Optional[List[int]] = None,
    ):
        m = max(len(math.atleast_1d(x)), len(math.atleast_1d(y)))
        super().__init__(
            modes=modes or list(range(m)),
            name="Dgate",
        )
        self._add_parameter(make_parameter(x_trainable, x, "x", x_bounds))
        self._add_parameter(make_parameter(y_trainable, y, "y", y_bounds))

    @property
    def d_vector(self):
        return gaussian.displacement(self.x.value, self.y.value)

    def U(self, cutoffs: Optional[Sequence[int]] = None, shape: Optional[Sequence[int]] = None):
        r"""Returns the unitary representation of the Displacement gate using
        the Laguerre polynomials.

        If specified, ``shape`` takes precedence over ``cutoffs``.
        ``shape`` is in the order ``(out, in)``.

        Note that for a unitary transformation on N modes, ``len(cutoffs)`` is ``N``
        and ``len(shape)`` is ``2N``.

        Arguments:
            cutoffs: the Hilbert space dimension cutoff for each mode.
            shape: the shape of the unitary matrix.

        Returns:
           Raises:
               ValueError: if the length of the cutoffs array is different from N and 2N
        """
        N = self.num_modes
        if cutoffs is None:
            pass
        elif len(cutoffs) == N:
            cutoffs = tuple(cutoffs) * 2
        elif len(cutoffs) == 2 * N:
            cutoffs = tuple(cutoffs)
        else:
            raise ValueError(
                "len(cutoffs) should be either equal to the number of modes or twice the number of modes (for output-input)."
            )
        shape = shape or cutoffs
        if shape is None:
            raise ValueError

        x = self.x.value * math.ones(N, dtype=self.x.value.dtype)
        y = self.y.value * math.ones(N, dtype=self.y.value.dtype)

        if N > 1:
            # calculate displacement unitary for each mode and concatenate with outer product
            Ud = None
            for idx, out_in in enumerate(zip(shape[:N], shape[N:])):
                if Ud is None:
                    Ud = fock.displacement(x[idx], y[idx], shape=out_in)
                else:
                    U_next = fock.displacement(x[idx], y[idx], shape=out_in)
                    Ud = math.outer(Ud, U_next)

            return math.transpose(
                Ud,
                list(range(0, 2 * N, 2)) + list(range(1, 2 * N, 2)),
            )
        else:
            return fock.displacement(x[0], y[0], shape=shape)


class Sgate(Unitary):
    r"""Squeezing gate.

    If ``len(modes) > 1`` the gate is applied in parallel to all of the modes provided.

    If a parameter is a single float, the parallel instances of the gate share that parameter.

    To apply mode-specific values use a list of floats. One can optionally set bounds for each
    parameter, which the optimizer will respect.

    Args:
        r (float or List[float]): the list of squeezing magnitudes
        r_bounds (float, float): bounds for the squeezing magnitudes
        r_trainable (bool): whether r is a trainable variable
        phi (float or List[float]): the list of squeezing angles
        phi_bounds (float, float): bounds for the squeezing angles
        phi_trainable bool: whether phi is a trainable variable
        modes (optional, List[int]): the list of modes this gate is applied to
    """

    is_gaussian = True
    short_name = "S"
    parallelizable = True

    def __init__(
        self,
        r: Union[float, list[float]] = 0.0,
        phi: Union[float, list[float]] = 0.0,
        r_trainable: bool = False,
        phi_trainable: bool = False,
        r_bounds: Tuple[Optional[float], Optional[float]] = (0.0, None),
        phi_bounds: Tuple[Optional[float], Optional[float]] = (None, None),
        modes: Optional[list[int]] = None,
    ):
        super().__init__(
            modes=modes or list(range(len(math.atleast_1d(r)))),  # type: ignore
            name="Sgate",
        )
        self._add_parameter(make_parameter(r_trainable, r, "r", r_bounds))
        self._add_parameter(make_parameter(phi_trainable, phi, "phi", phi_bounds))

    def U(self, cutoffs: Optional[Sequence[int]] = None, shape: Optional[Sequence[int]] = None):
        r"""Returns the unitary representation of the Squeezing gate.

        If specified, ``shape`` takes precedence over ``cutoffs``.
        ``shape`` is in the order ``(out, in)``.

        Note that for a unitary transformation on N modes, ``len(cutoffs)`` is ``N``
        and ``len(shape)`` is ``2N``.

        Args:
            cutoffs: the Hilbert space dimension cutoff for each mode.
            shape: the shape of the unitary matrix.

        Returns:
            array[complex]: the unitary matrix
        """
        N = self.num_modes
        if cutoffs is None:
            pass
        elif len(cutoffs) == N:
            cutoffs = tuple(cutoffs) * 2
        elif len(cutoffs) == 2 * N:
            cutoffs = tuple(cutoffs)
        else:
            raise ValueError(
                "len(cutoffs) should be either equal to the number of modes or twice the number of modes (for output-input)."
            )
        shape = shape or cutoffs
        if shape is None:
            raise ValueError

        # this works both or scalar r/phi and vector r/phi:
        r = self.r.value * math.ones(N, dtype=self.r.value.dtype)
        phi = self.phi.value * math.ones(N, dtype=self.phi.value.dtype)

        if N > 1:
            # calculate squeezing unitary for each mode and concatenate with outer product
            Us = None
            for idx, single_shape in enumerate(zip(shape[:N], shape[N:])):
                if Us is None:
                    Us = fock.squeezer(r[idx], phi[idx], shape=single_shape)
                else:
                    U_next = fock.squeezer(r[idx], phi[idx], shape=single_shape)
                    Us = math.outer(Us, U_next)
            return math.transpose(
                Us,
                list(range(0, 2 * N, 2)) + list(range(1, 2 * N, 2)),
            )
        else:
            return fock.squeezer(r[0], phi[0], shape=shape)

    @property
    def X_matrix(self):
        return gaussian.squeezing_symplectic(self.r.value, self.phi.value)


class Rgate(Unitary):
    r"""Rotation gate.

    If ``len(modes) > 1`` the gate is applied in parallel to all of the modes provided.

    If a parameter is a single float, the parallel instances of the gate share that parameter.

    To apply mode-specific values use a list of floats. One can optionally set bounds for each
    parameter, which the optimizer will respect.

    Args:
        modes (List[int]): the list of modes this gate is applied to
        angle (float or List[float]): the list of rotation angles
        angle_bounds (float, float): bounds for the rotation angles
        angle_trainable bool: whether angle is a trainable variable
        modes (optional, List[int]): the list of modes this gate is applied to
    """

    is_gaussian = True
    short_name = "R"
    parallelizable = True

    def __init__(
        self,
        angle: Union[float, list[float]] = 0.0,
        angle_trainable: bool = False,
        angle_bounds: Tuple[Optional[float], Optional[float]] = (None, None),
        modes: Optional[list[int]] = None,
    ):
        super().__init__(
            modes=modes or list(range(len(math.atleast_1d(angle)))),  # type: ignore
            name="Rgate",
        )
        self._add_parameter(make_parameter(angle_trainable, angle, "angle", angle_bounds))

    @property
    def X_matrix(self):
        return gaussian.rotation_symplectic(self.angle.value)

    def U(
        self,
        cutoffs: Optional[Sequence[int]] = None,
        shape: Optional[Sequence[int]] = None,
        diag_only=False,
    ):
        r"""Returns the unitary representation of the Rotation gate.

        If specified, ``shape`` takes precedence over ``cutoffs``.
        ``shape`` is in the order ``(out, in)``.

        Note that for a unitary transformation on N modes, ``len(cutoffs)`` is ``N``
        and ``len(shape)`` is ``2N``.

        Args:
            cutoffs: cutoff dimension for each mode.
            shape: the shape of the unitary matrix
            diag_only: if True, only return the diagonal of the unitary matrix.

        Returns:
            array[complex]: the unitary matrix
        """
        N = self.num_modes
        if diag_only:
            raise NotImplementedError("Rgate does not support diag_only=True yet")
        if cutoffs is None:
            pass
        elif len(cutoffs) == N:
            cutoffs = tuple(cutoffs) * 2
        elif len(cutoffs) == 2 * N:
            cutoffs = tuple(cutoffs)
        else:
            raise ValueError(
                "len(cutoffs) should be either equal to the number of modes or twice the number of modes (for output-input)."
            )
        shape = shape or cutoffs
        if shape is None:
            raise ValueError

        angles = self.angle.value * math.ones(self.num_modes, dtype=self.angle.value.dtype)

        # calculate rotation unitary for each mode and concatenate with outer product
        Ur = None
        for idx, cutoff in enumerate(shape[: self.num_modes]):
            theta = math.arange(cutoff) * angles[idx]
            if Ur is None:
                Ur = math.diag(math.make_complex(math.cos(theta), math.sin(theta)))
            else:
                U_next = math.diag(math.make_complex(math.cos(theta), math.sin(theta)))
                Ur = math.outer(Ur, U_next)

        # return total unitary with indexes reordered according to MM convention
        return math.transpose(
            Ur,
            list(range(0, 2 * self.num_modes, 2)) + list(range(1, 2 * self.num_modes, 2)),
        )


class Pgate(Unitary):
    r"""Quadratic phase gate.

    If len(modes) > 1 the gate is applied in parallel to all of the modes provided. If a parameter
    is a single float, the parallel instances of the gate share that parameter. To apply
    mode-specific values use a list of floats. One can optionally set bounds for each parameter,
    which the optimizer will respect.

    Args:
        modes (List[int]): the list of modes this gate is applied to
        shearing (float or List[float]): the list of shearing parameters
        shearing_bounds (float, float): bounds for the shearing parameters
        shearing_trainable bool: whether shearing is a trainable variable
        modes (optional, List[int]): the list of modes this gate is applied to
    """

    is_gaussian = True
    short_name = "P"
    parallelizable = True

    def __init__(
        self,
        shearing: Union[Optional[float], Optional[list[float]]] = 0.0,
        shearing_trainable: bool = False,
        shearing_bounds: Tuple[Optional[float], Optional[float]] = (None, None),
        modes: Optional[list[int]] = None,
    ):
        super().__init__(
            modes=modes or list(range(len(math.atleast_1d(shearing)))),
            name="Pgate",
        )
        self._add_parameter(
            make_parameter(shearing_trainable, shearing, "shearing", shearing_bounds)
        )

    @property
    def X_matrix(self):
        return gaussian.quadratic_phase(self.shearing.value)


class CXgate(Unitary):
    r"""Controlled X gate.

    It applies to a single pair of modes. One can optionally set bounds for each parameter, which
    the optimizer will respect.

    Args:
        s (float): control parameter
        s_bounds (float, float): bounds for the control angle
        s_trainable (bool): whether s is a trainable variable
        modes (optional, List[int]): the list of modes this gate is applied to
    """

    is_gaussian = True
    short_name = "CX"
    parallelizable = False

    def __init__(
        self,
        s: Optional[float] = 0.0,
        s_trainable: bool = False,
        s_bounds: Tuple[Optional[float], Optional[float]] = (None, None),
        modes: Optional[List[int]] = None,
    ):
        super().__init__(
            modes=modes or [0, 1],
            name="CXgate",
        )
        self._add_parameter(make_parameter(s_trainable, s, "s", s_bounds))

    @property
    def X_matrix(self):
        return gaussian.controlled_X(self.s.value)


class CZgate(Unitary):
    r"""Controlled Z gate.

    It applies to a single pair of modes. One can optionally set bounds for each parameter, which
    the optimizer will respect.

    Args:
        s (float): control parameter
        s_bounds (float, float): bounds for the control angle
        s_trainable (bool): whether s is a trainable variable
        modes (optional, List[int]): the list of modes this gate is applied to
    """

    is_gaussian = True
    short_name = "CZ"
    parallelizable = False

    def __init__(
        self,
        s: Optional[float] = 0.0,
        s_trainable: bool = False,
        s_bounds: Tuple[Optional[float], Optional[float]] = (None, None),
        modes: Optional[List[int]] = None,
    ):
        super().__init__(
            modes=modes or [0, 1],
            name="CZgate",
        )
        self._add_parameter(make_parameter(s_trainable, s, "s", s_bounds))

    @property
    def X_matrix(self):
        return gaussian.controlled_Z(self.s.value)


class BSgate(Unitary):
    r"""Beam splitter gate.

    It applies to a single pair of modes.
    One can optionally set bounds for each parameter, which the optimizer will respect.

    Args:
        theta (float): the transmissivity angle
        theta_bounds (float, float): bounds for the transmissivity angle
        theta_trainable (bool): whether theta is a trainable variable
        phi (float): the phase angle
        phi_bounds (float, float): bounds for the phase angle
        phi_trainable bool: whether phi is a trainable variable
        modes (optional, List[int]): the list of modes this gate is applied to
    """

    is_gaussian = True
    short_name = "BS"
    parallelizable = False

    def __init__(
        self,
        theta: float = 0.0,
        phi: float = 0.0,
        theta_trainable: bool = False,
        phi_trainable: bool = False,
        theta_bounds: Tuple[Optional[float], Optional[float]] = (None, None),
        phi_bounds: Tuple[Optional[float], Optional[float]] = (None, None),
        modes: Optional[list[int]] = None,
    ):
        super().__init__(
            modes=modes or [0, 1],  # type: ignore
            name="BSgate",
        )
        self._add_parameter(make_parameter(theta_trainable, theta, "theta", theta_bounds))
        self._add_parameter(make_parameter(phi_trainable, phi, "phi", phi_bounds))

    def U(
        self,
        cutoffs: Optional[List[int]] = None,
        shape: Optional[Sequence[int]] = None,
        method=None,
    ):
        r"""Returns the unitary representation of the beam splitter.

        If specified, ``shape`` takes precedence over ``cutoffs``.
        ``shape`` is in the order ``(out, in)``.

        Note that for a unitary transformation on N modes, ``len(cutoffs)`` is ``N``
        and ``len(shape)`` is ``2N``.

        Args:
            cutoffs: the list of cutoff dimensions for each mode
                in the order (out_0, out_1, in_0, in_1).
            shape: the shape of the unitary matrix
            method: the method used to compute the unitary matrix. Options are:
                * 'vanilla': uses the standard method
                * 'schwinger': slower, but numerically stable
            default is set in settings.DEFAULT_BS_METHOD (with 'vanilla' by default)

        Returns:
            array[complex]: the unitary tensor of the beamsplitter
        """
        if cutoffs is None:
            pass
        elif len(cutoffs) == 4:
            shape = tuple(cutoffs)
        elif len(cutoffs) == 2:
            shape = tuple(cutoffs) + tuple(cutoffs)
        else:
            raise ValueError(f"Invalid len(cutoffs): {len(cutoffs)} (should be 2 or 4).")

        shape = shape or cutoffs

        return fock.beamsplitter(
            self.theta.value,
            self.phi.value,
            shape=shape,
            method=method or settings.DEFAULT_BS_METHOD,
        )

    @property
    def X_matrix(self):
        return gaussian.beam_splitter_symplectic(self.theta.value, self.phi.value)

    def _validate_modes(self, modes):
        if len(modes) != 2:
            raise ValueError(
                f"Invalid number of modes: {len(modes)} (should be 2). Perhaps you are looking for Interferometer."
            )


class MZgate(Unitary):
    r"""Mach-Zehnder gate.

    It supports two conventions:
        1. if ``internal=True``, both phases act inside the interferometer: ``phi_a`` on the upper arm, ``phi_b`` on the lower arm;
        2. if ``internal = False``, both phases act on the upper arm: ``phi_a`` before the first BS, ``phi_b`` after the first BS.

    One can optionally set bounds for each parameter, which the optimizer will respect.

    Args:
        phi_a (float): the phase in the upper arm of the MZ interferometer
        phi_a_bounds (float, float): bounds for phi_a
        phi_a_trainable (bool): whether phi_a is a trainable variable
        phi_b (float): the phase in the lower arm or external of the MZ interferometer
        phi_b_bounds (float, float): bounds for phi_b
        phi_b_trainable (bool): whether phi_b is a trainable variable
        internal (bool): whether phases are both in the internal arms (default is False)
        modes (optional, List[int]): the list of modes this gate is applied to
    """

    is_gaussian = True
    short_name = "MZ"
    parallelizable = False

    def __init__(
        self,
        phi_a: float = 0.0,
        phi_b: float = 0.0,
        phi_a_trainable: bool = False,
        phi_b_trainable: bool = False,
        phi_a_bounds: Tuple[Optional[float], Optional[float]] = (None, None),
        phi_b_bounds: Tuple[Optional[float], Optional[float]] = (None, None),
        internal: bool = False,
        modes: Optional[List[int]] = None,
    ):
        super().__init__(
            modes=modes or [0, 1],
            name="MZgate",
        )
        self._add_parameter(make_parameter(phi_a_trainable, phi_a, "phi_a", phi_a_bounds))
        self._add_parameter(make_parameter(phi_b_trainable, phi_b, "phi_b", phi_b_bounds))
        self._internal = internal

    @property
    def X_matrix(self):
        return gaussian.mz_symplectic(self.phi_a.value, self.phi_b.value, internal=self._internal)

    def _validate_modes(self, modes):
        if len(modes) != 2:
            raise ValueError(
                f"Invalid number of modes: {len(modes)} (should be 2). Perhaps you are looking for Interferometer?"
            )


class S2gate(Unitary):
    r"""Two-mode squeezing gate.

    It applies to a single pair of modes. One can optionally set bounds for each parameter, which the optimizer will respect.

    Args:
        r (float): the squeezing magnitude
        r_bounds (float, float): bounds for the squeezing magnitude
        r_trainable (bool): whether r is a trainable variable
        phi (float): the squeezing angle
        phi_bounds (float, float): bounds for the squeezing angle
        phi_trainable bool: whether phi is a trainable variable
        modes (optional, List[int]): the list of modes this gate is applied to
    """

    is_gaussian = True
    short_name = "S2"
    parallelizable = False

    def __init__(
        self,
        r: float = 0.0,
        phi: float = 0.0,
        r_trainable: bool = False,
        phi_trainable: bool = False,
        r_bounds: Tuple[Optional[float], Optional[float]] = (0.0, None),
        phi_bounds: Tuple[Optional[float], Optional[float]] = (None, None),
        modes: Optional[List[int]] = None,
    ):
        super().__init__(
            modes=modes or [0, 1],
            name="S2gate",
        )
        self._add_parameter(make_parameter(r_trainable, r, "r", r_bounds))
        self._add_parameter(make_parameter(phi_trainable, phi, "phi", phi_bounds))

    @property
    def X_matrix(self):
        return gaussian.two_mode_squeezing_symplectic(self.r.value, self.phi.value)

    def _validate_modes(self, modes):
        if len(modes) != 2:
            raise ValueError(f"Invalid number of modes: {len(modes)} (should be 2")


class Interferometer(Unitary):
    r"""N-mode interferometer.

    It corresponds to a Ggate with zero mean and a ``2N x 2N`` unitary symplectic matrix.

    Args:
        num_modes (int): the num_modes-mode interferometer
        unitary (2d array): a valid unitary matrix U. For N modes it must have shape `(N,N)`
        unitary_trainable (bool): whether unitary is a trainable variable
        modes (optional, List[int]): the list of modes this gate is applied to
    """

    is_gaussian = True
    short_name = "I"
    parallelizable = False

    def __init__(
        self,
        num_modes: int,
        unitary: Optional[ComplexMatrix] = None,
        unitary_trainable: bool = False,
        modes: Optional[list[int]] = None,
    ):
        if modes is not None and num_modes != len(modes):
            raise ValueError(f"Invalid number of modes: got {len(modes)}, should be {num_modes}")
        if unitary is None:
            unitary = math.random_unitary(num_modes)
        super().__init__(
            modes=modes or list(range(num_modes)),
            name="Interferometer",
        )
        self._add_parameter(
            make_parameter(unitary_trainable, unitary, "unitary", (None, None), update_unitary)
        )

    @property
    def X_matrix(self):
        return math.block(
            [
                [math.real(self.unitary.value), -math.imag(self.unitary.value)],
                [math.imag(self.unitary.value), math.real(self.unitary.value)],
            ]
        )

    def _validate_modes(self, modes):
        if len(modes) != self.unitary.value.shape[-1]:
            raise ValueError(
                f"Invalid number of modes: {len(modes)} (should be {self.unitary.shape[-1]})"
            )

    def __repr__(self):
        modes = self.modes
        unitary = repr(math.asnumpy(self.unitary.value)).replace("\n", "")
        return f"Interferometer(num_modes = {len(modes)}, unitary = {unitary}){modes}"


class RealInterferometer(Unitary):
    r"""N-mode interferometer parametrized by an NxN orthogonal matrix (or 2N x 2N block-diagonal orthogonal matrix). This interferometer does not mix q and p.
    Does not mix q's and p's.

    Args:
        orthogonal (2d array, optional): a real unitary (orthogonal) matrix. For N modes it must have shape `(N,N)`.
            If set to `None` a random real unitary (orthogonal) matrix is used.
        orthogonal_trainable (bool): whether orthogonal is a trainable variable
    """

    is_gaussian = True
    short_name = "RI"
    parallelizable = False

    def __init__(
        self,
        num_modes: int,
        orthogonal: Optional[RealMatrix] = None,
        orthogonal_trainable: bool = False,
        modes: Optional[List[int]] = None,
    ):
        if modes is not None and (num_modes != len(modes)):
            raise ValueError(f"Invalid number of modes: got {len(modes)}, should be {num_modes}")
        if orthogonal is None:
            orthogonal = math.random_orthogonal(num_modes)

        super().__init__(
            modes=modes or list(range(num_modes)),
            name="RealInterferometer",
        )
        self._add_parameter(
            make_parameter(
                orthogonal_trainable, orthogonal, "orthogonal", (None, None), update_orthogonal
            )
        )

    @property
    def X_matrix(self):
        return math.block(
            [
                [self.orthogonal.value, -math.zeros_like(self.orthogonal.value)],
                [math.zeros_like(self.orthogonal.value), self.orthogonal.value],
            ]
        )

    def _validate_modes(self, modes):
        if len(modes) != self.orthogonal.value.shape[-1]:
            raise ValueError(
                f"Invalid number of modes: {len(modes)} (should be {self.orthogonal.value.shape[-1]})"
            )

    def __repr__(self):
        modes = self.modes
        orthogonal = repr(math.asnumpy(self.orthogonal.value)).replace("\n", "")
        return f"RealInterferometer(num_modes = {len(modes)}, orthogonal = {orthogonal}){modes}"


class Ggate(Unitary):
    r"""A generic N-mode Gaussian unitary transformation with zero displacement.

    If a symplectic matrix is not provided, one will be picked at random with effective squeezing
    strength ``r`` in ``[0, 1]`` for each mode.

    Args:
        num_modes (int): the number of modes this gate is acting on.
        symplectic (2d array): a valid symplectic matrix in XXPP order. For N modes it must have shape ``(2N,2N)``.
        symplectic_trainable (bool): whether symplectic is a trainable variable.
    """

    is_gaussian = True
    short_name = "G"
    parallelizable = False

    def __init__(
        self,
        num_modes: int,
        symplectic: Optional[RealMatrix] = None,
        symplectic_trainable: bool = False,
        modes: Optional[list[int]] = None,
    ):
        if modes is not None and (num_modes != len(modes)):
            raise ValueError(f"Invalid number of modes: got {len(modes)}, should be {num_modes}")
        if symplectic is None:
            symplectic = math.random_symplectic(num_modes)

        super().__init__(
            modes=modes or list(range(num_modes)),
            name="Ggate",
        )
        self._add_parameter(
            make_parameter(
                symplectic_trainable, symplectic, "symplectic", (None, None), update_symplectic
            )
        )

    @property
    def X_matrix(self):
        return self.symplectic.value

    def _validate_modes(self, modes):
        if len(modes) != self.symplectic.value.shape[1] // 2:
            raise ValueError(
                f"Invalid number of modes: {len(modes)} (should be {self.symplectic.value.shape[1] // 2})"
            )

    def __repr__(self):
        modes = self.modes
        symplectic = repr(math.asnumpy(self.symplectic.value)).replace("\n", "")
        return f"Ggate(num_modes = {len(modes)}, symplectic = {symplectic}){modes}"


# ~~~~~~~~~~~~~
# NON-UNITARY
# ~~~~~~~~~~~~~


# pylint: disable=no-member
class Attenuator(Channel):
    r"""The noisy attenuator channel.

    It corresponds to mixing with a thermal environment and applying the pure loss channel. The pure
    lossy channel is recovered for nbar = 0 (i.e. mixing with vacuum).

    The CPT channel is given by

    .. math::

        X = sqrt(transmissivity) * I
        Y = (1-transmissivity) * (2*nbar + 1) * (hbar / 2) * I

    If ``len(modes) > 1`` the gate is applied in parallel to all of the modes provided.
    If ``transmissivity`` is a single float, the parallel instances of the gate share that parameter.

    To apply mode-specific values use a list of floats.

    One can optionally set bounds for `transmissivity`, which the optimizer will respect.

    Args:
        transmissivity (float or List[float]): the list of transmissivities
        nbar (float): the average number of photons in the thermal state
        transmissivity_trainable (bool): whether transmissivity is a trainable variable
        nbar_trainable (bool): whether nbar is a trainable variable
        transmissivity_bounds (float, float): bounds for the transmissivity
        nbar_bounds (float, float): bounds for the average number of photons in the thermal state
        modes (optional, List[int]): the list of modes this gate is applied to
    """

    is_gaussian = True
    short_name = "Att"
    parallelizable = True

    def __init__(
        self,
        transmissivity: Union[Optional[float], Optional[List[float]]] = 1.0,
        nbar: float = 0.0,
        transmissivity_trainable: bool = False,
        nbar_trainable: bool = False,
        transmissivity_bounds: Tuple[Optional[float], Optional[float]] = (0.0, 1.0),
        nbar_bounds: Tuple[Optional[float], Optional[float]] = (0.0, None),
        modes: Optional[List[int]] = None,
    ):
        super().__init__(
            modes=modes or list(range(len(math.atleast_1d(transmissivity)))),
            name="Attenuator",
        )
        self._add_parameter(
            make_parameter(
                transmissivity_trainable,
                transmissivity,
                "transmissivity",
                transmissivity_bounds,
                None,
            )
        )
        self._add_parameter(make_parameter(nbar_trainable, nbar, "nbar", nbar_bounds))

    @property
    def X_matrix(self):
        return gaussian.loss_XYd(self.transmissivity.value, self.nbar.value)[0]

    @property
    def Y_matrix(self):
        return gaussian.loss_XYd(self.transmissivity.value, self.nbar.value)[1]


class Amplifier(Channel):
    r"""The noisy amplifier channel.

    It corresponds to mixing with a thermal environment and applying a two-mode squeezing gate.

    .. code:: python

        X = sqrt(gain) * I
        Y = (gain-1) * (2*nbar + 1) * (hbar / 2) * I

    If ``len(modes) > 1`` the gate is applied in parallel to all of the modes provided.
    If ``gain`` is a single float, the parallel instances of the gate share that parameter.
    To apply mode-specific values use a list of floats.
    One can optionally set bounds for ``gain``, which the optimizer will respect.

    Args:
        gain (float or List[float]): the list of gains (must be > 1)
        nbar (float): the average number of photons in the thermal state
        nbar_trainable (bool): whether nbar is a trainable variable
        gain_trainable (bool): whether gain is a trainable variable
        gain_bounds (float, float): bounds for the gain
        nbar_bounds (float, float): bounds for the average number of photons in the thermal state
        modes (optional, List[int]): the list of modes this gate is applied to
    """

    is_gaussian = True
    short_name = "Amp"
    parallelizable = True

    def __init__(
        self,
        gain: Union[Optional[float], Optional[List[float]]] = 1.0,
        nbar: float = 0.0,
        gain_trainable: bool = False,
        nbar_trainable: bool = False,
        gain_bounds: Tuple[Optional[float], Optional[float]] = (1.0, None),
        nbar_bounds: Tuple[Optional[float], Optional[float]] = (0.0, None),
        modes: Optional[list[int]] = None,
    ):
        super().__init__(
            modes=modes or list(range(len(math.atleast_1d(gain)))),
            name="Amplifier",
        )
        self._add_parameter(make_parameter(gain_trainable, gain, "gain", gain_bounds))
        self._add_parameter(make_parameter(nbar_trainable, nbar, "nbar", nbar_bounds))

    @property
    def X_matrix(self):
        return gaussian.amp_XYd(self.gain.value, self.nbar.value)[0]

    @property
    def Y_matrix(self):
        return gaussian.amp_XYd(self.gain.value, self.nbar.value)[1]


# pylint: disable=no-member
class AdditiveNoise(Channel):
    r"""The additive noise channel.

    Equivalent to an amplifier followed by an attenuator. E.g.,

    .. code-block::

        na,nb = np.random.uniform(size=2)
        tr = np.random.uniform()
        Amplifier(1/tr, nb) >> Attenuator(tr, na) == AdditiveNoise(2*(1-tr)*(1+na+nb)) # evaluates to True

    or equivalent to an attenuator followed by an amplifier:

    .. code-block::

        na,nb = np.random.uniform(size=2)
        amp = 1.0 + np.random.uniform()
        Attenuator(1/amp, nb) >> Amplifier(amp, na) == AdditiveNoise(2*(amp-1)*(1+na+nb))

    Args:
        noise (float or List[float]): the added noise in units of hbar/2
        noise_trainable (bool): whether noise is a trainable variable
        noise_bounds (float, float): bounds for the noise
        modes (optional, List[int]): the list of modes this gate is applied to
    """

    is_gaussian = True
    short_name = "Add"
    parallelizable = True

    def __init__(
        self,
        noise: Union[float, list[float]] = 0.0,
        noise_trainable: bool = False,
        noise_bounds: Tuple[Optional[float], Optional[float]] = (0.0, None),
        modes: Optional[list[int]] = None,
    ):
        super().__init__(
            modes=modes or list(range(len(math.atleast_1d(noise)))),
            name="AddNoise",
        )
        self._add_parameter(make_parameter(noise_trainable, noise, "noise", noise_bounds))

    @property
    def Y_matrix(self):
        return gaussian.noise_Y(self.noise.value)


class PhaseNoise(Channel):
    r"""The phase noise channel.

    The phase noise channel is a non-Gaussian transformation that is equivalent to
    a random phase rotation.

    Args:
        phase_stdev (float or List[float]): the standard deviation of the (wrapped) normal
            distribution for the angle of the rotation
        modes (optional, list(int)): the single mode this gate is applied to (default [0])
    """

    def __init__(
        self,
        phase_stdev: Union[Optional[float], Optional[List[float]]] = 0.0,
        phase_stdev_trainable: bool = False,
        phase_stdev_bounds: Tuple[Optional[float], Optional[float]] = (0.0, None),
        modes: Optional[List[int]] = None,
    ):
        super().__init__(
            modes=modes or [0],
            name="AddNoise",
        )
        self._add_parameter(
            make_parameter(phase_stdev_trainable, phase_stdev, "phase_stdev", phase_stdev_bounds)
        )

        self._modes = modes or [0]
        self.is_unitary = False
        self.is_gaussian = False
        self.short_name = "P~"

    # need to override primal because of the unconventional way
    # the channel is defined in Fock representation
    def primal(self, state):
        idx = state.modes.index(self.modes[0])
        if state.is_pure:
            ket = state.ket()
            dm = fock.ket_to_dm(ket)
        else:
            dm = state.dm()

        # transpose dm so that the modes of interest are at the end
        M = state.num_modes
        indices = list(range(2 * M))
        indices.remove(idx)
        indices.remove(idx + M)
        indices += [idx, idx + M]
        dm = math.transpose(dm, indices)

        coeff = math.cast(
            math.exp(
                -0.5 * self.phase_stdev.value**2 * math.arange(-dm.shape[-2] + 1, dm.shape[-1]) ** 2
            ),
            dm.dtype,
        )

        for k in range(-dm.shape[-2] + 1, dm.shape[-1]):
            diagonal = math.diag_part(dm, k=k)
            diagonal *= coeff[k + dm.shape[-2] - 1]
            dm = math.set_diag(dm, diagonal, k=k)

        # transpose dm back to the original order
        return State(dm=math.transpose(dm, np.argsort(indices)), modes=state.modes)



# Copyright 2023 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


# Recurrencies for Fock-Bargmann amplitudes put together a strategy for
# enumerating the indices in a specific order and functions for calculating
# which neighbours to use in the calculation of the amplitude at a given index.
# In summary, they return the value of the amplitude at the given index by following
# a recipe made of two parts. The function to recompute A and b is determined by
# which neighbours are used.

"""Fock-Bargmann recurrence relation steps optimized for beamsplitter."""

import numpy as np
from numba import njit

from mrmustard.math.lattice import steps
from mrmustard.utils.typing import ComplexMatrix, ComplexTensor, ComplexVector

SQRT = np.sqrt(np.arange(100000))

__all__ = ["beamsplitter", "beamsplitter_vjp", "beamsplitter_schwinger"]


@njit
def beamsplitter(
    shape: tuple[int, int, int, int], theta: float, phi: float, dtype=np.complex128
) -> ComplexTensor:  # pragma: no cover
    r"""Calculates the Fock representation of the beamsplitter.
    It takes advantage of input-output particle conservation (m+n=p+q)
    to avoid one for loop. Inspired from the original implementation in
    the walrus by @ziofil. Here is how the parameters are used in the
    code (see eq. 73-75 in https://arxiv.org/abs/2004.11002):

    A = [[0,V],[v^T,0]]   # BS bargmann matrix
    V = [[ct, -st expm],  # BS unitary
         [st exp, ct]]

    Args:
        shape (tuple[int, int, int, int]): shape of the Fock representation
        theta (float): beamsplitter angle
        phi (float): beamsplitter phase
        dtype (np.dtype): data type of the Fock representation

    Returns:
        array (ComplexTensor): The Fock representation of the gate
    """
    ct = np.cos(theta)
    st = np.sin(theta) * np.exp(1j * phi)
    stc = np.conj(st)

    M, N, P, Q = shape
    G = np.zeros(shape, dtype=dtype)
    G[0, 0, 0, 0] = 1.0

    # rank 3
    for m in range(M):
        for n in range(N - m):
            p = m + n
            if 0 < p < P:
                G[m, n, p, 0] = (
                    ct * SQRT[m] / SQRT[p] * G[m - 1, n, p - 1, 0]
                    + st * SQRT[n] / SQRT[p] * G[m, n - 1, p - 1, 0]
                )

    # rank 4
    for m in range(M):
        for n in range(N):
            for p in range(P):
                q = m + n - p
                if 0 < q < Q:
                    G[m, n, p, q] = (
                        -stc * SQRT[m] / SQRT[q] * G[m - 1, n, p, q - 1]
                        + ct * SQRT[n] / SQRT[q] * G[m, n - 1, p, q - 1]
                    )
    return G


@njit
def beamsplitter_vjp(
    G: ComplexTensor,
    dLdG: ComplexTensor,
    theta: float,
    phi: float,
) -> tuple[ComplexMatrix, ComplexVector, complex]:  # pragma: no cover
    r"""Beamsplitter gradients with respect to theta and phi.
    This function could return dL/dA, dL/db, dL/dc like its vanilla counterpart,
    but it is more efficient to include this chain rule step in the numba function,
    since we can.

    We use these derivatives of the BS unitary:

    dVdt = [[-st, -ct exp],
            [ct exp, -st]]
    dVdphi = [[0, -i ct expm],
              [i ct exp, 0]]

    Args:
        G (np.ndarray): Tensor result of the forward pass
        dLdG (np.ndarray): gradient of the loss with respect to the output tensor
        theta (float): beamsplitter angle
        phi (float): beamsplitter phase

    Returns:
        tuple[float, float]: dL/dtheta, dL/dphi
    """
    M, N, P, Q = G.shape

    # init gradients
    dA = np.zeros((4, 4), dtype=np.complex128)
    db = np.zeros(4, dtype=np.complex128)
    dLdA = np.zeros_like(dA)
    dLdb = np.zeros_like(db)

    # rank 3
    for m in range(M):
        for n in range(N - m):
            p = m + n
            if 0 < p < P:
                dA, db = steps.vanilla_step_grad(G, (m, n, p, 0), dA, db)
                dLdA += dA * dLdG[m, n, p, 0]
                dLdb += db * dLdG[m, n, p, 0]

    # rank 4
    for m in range(M):
        for n in range(N):
            for p in range(P):
                q = m + n - p
                if 0 < q < Q:
                    dA, db = steps.vanilla_step_grad(G, (m, n, p, q), dA, db)
                    dLdA += dA * dLdG[m, n, p, q]
                    dLdb += db * dLdG[m, n, p, q]

    st = np.sin(theta)
    ct = np.cos(theta)
    e = np.exp(1j * phi)
    em = np.exp(-1j * phi)

    # omitting bottom-left block because dLdA should be zero there
    dLdtheta = 2 * np.real(
        -st * dLdA[0, 2] - ct * em * dLdA[0, 3] + ct * e * dLdA[1, 2] - st * dLdA[1, 3]
    )
    dLdphi = 2 * np.real(1j * st * em * dLdA[0, 3] + 1j * st * e * dLdA[1, 2])

    return dLdtheta, dLdphi


def beamsplitter_schwinger(shape, theta, phi, max_N=None):
    r"""Returns the Fock representation of the beamsplitter up to
    the given cutoff for each of the two modes.

    This implementation is in pure python (so it's slower than the numba version),
    but it's numerically stable up to arbitrary cutoffs.

    In this implementation we split the two-mode Fock basis into finite subsets spanned
    by |m,n> with m+n=const, i.e. {|0,0>}, {|1,0>,|0,1>}, {|2,0>, |1,1>, |0,2>}, etc...
    A beamsplitter acts unitariliy in each of these subspaces without mixing them with each other,
    i.e. in this basis the beamsplitter would be a block-diagonal matrix.
    This means we can construct the BS matrix by first calculating the BS unitaries
    in each of these subspaces.
    This can be done using the matrix exponential, which is numerically stable.
    We couldn't do this in the original basis because it was infinite-dimensional
    and we had to truncate it at some point.

    Arguments:
        shape (int, int, int, int): The shape of the output tensor. Only shapes of the form (i,k,i,k) are supported.
        theta (float): The angle of the beamsplitter.
        phi (float): The phase of the beamsplitter.
        max_N (int): The maximum total photon number to include in the calculation.

    Returns:
        np.ndarray: The beamsplitter in the Fock basis.
    """
    c1, c2, c3, c4 = shape
    if c1 != c3 or c2 != c4:
        raise ValueError("The Schwinger method only supports shapes of the form (i,k,i,k).")
    # create output tensor
    U = np.zeros(shape, dtype="complex128")

    # loop over subspaces of constant photon number N up to max_N
    if max_N is None or max_N > c1 + c2 - 2:
        max_N = c1 + c2 - 2
    for N in range(max_N + 1):
        # construct the N+1 x N+1 unitary for this subspace
        diag = np.exp(1j * phi) * np.sqrt(np.arange(N, 0, -1) * np.arange(1, N + 1, 1))
        iJy = np.diag(diag, k=-1) - np.diag(np.conj(diag), k=1)
        E, V = np.linalg.eig(theta * iJy)
        block = V @ np.diag(np.exp(E)) @ np.conj(V.T)
        # insert the elements of the block into the output tensor
        for i in range(max(0, N + 1 - c1), min(N + 1, c1)):
            for j in range(max(0, N + 1 - c1), min(N + 1, c2)):
                U[N - i, i, N - j, j] = block[i, j]
    return U



# Copyright 2023 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

r"""
Contains the functions to operate with flattened indices.

Given a multi-dimensional ``np.ndarray``, we can index its elements using ``np.ndindex``.
Alternatevely, we can flatten the multi-dimensional array and index its elements with
``int``s (hereby referred to as ''flat indices''). 
"""

from typing import Iterator, Sequence
from numba import njit

import numpy as np


@njit
def first_available_pivot(
    index: int, strides: Sequence[int]
) -> tuple[int, tuple[int, ...]]:  # pragma: no cover
    r"""
    Returns the first available pivot for the given flat index.
    A pivot is a nearest neighbor of the index. Here we pick the first available pivot.

    Arguments:
        index: the flat index to get the first available pivot of.
        strides: the strides that allow mapping the flat index to a tuple index.

    Returns:
        the flat index that was decremented and the pivot.
    """
    for i, s in enumerate(strides):
        y = index - s
        if y >= 0:
            return (i, y)
    raise ValueError("Index is zero.")


@njit
def lower_neighbors(
    index: int, strides: Sequence[int], start: int
) -> Iterator[tuple[int, tuple[int, ...]]]:  # pragma: no cover
    r"""
    Yields the flat indices of the lower neighbours of the given flat index.
    """
    for i in range(start, len(strides)):
        yield i, index - strides[i]


@njit
def shape_to_strides(shape: Sequence[int]) -> Sequence[int]:  # pragma: no cover
    r"""
    Calculates strides from shape.

    Arguments:
        shape: the shape of the ``np.ndindex``.

    Returns:
        the strides that allow mapping a flat index to the corresponding ``np.ndindex``.
    """
    strides = np.ones_like(shape)
    for i in range(1, len(shape)):
        strides[i - 1] = np.prod(shape[i:])
    return strides



# Copyright 2023 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" Classes for constructing tensors."""

from __future__ import annotations

from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from typing import List, Optional, Tuple

import uuid

from mrmustard.math.backend_manager import BackendManager

math = BackendManager()


__all__ = ["Wire", "Tensor"]


def random_int() -> int:
    r"""
    Returns a random integer obtained from a UUID
    """
    return uuid.uuid1().int


# pylint: disable=too-many-instance-attributes
@dataclass
class Wire:
    r"""Represents a wire in a tensor network.

    Each wire is characterized by a unique identifier ``id``, which must be different from
    the identifiers of all the other wires in the tensor network. Additionally, it owns a
    label ``mode`` that represents the mode of light that this wire is acting on.

    Args:
        id: A numerical identifier for this wire.
        mode: The mode of light that this wire is acting on.
        is_input: Whether this wire is an input to a tensor or an output.
        is_ket: Whether this wire is on the ket or on the bra side.

    """

    id: int
    mode: int
    is_input: bool
    is_ket: bool

    def __post_init__(self):
        self._contraction_id: int = random_int()
        self._dim = None
        self._is_connected = False

    @property
    def contraction_id(self) -> int:
        r"""
        A numerical identifier for the contraction involving this wire.
        """
        return self._contraction_id

    @contraction_id.setter
    def contraction_id(self, value: int):
        self._contraction_id = value

    @property
    def dim(self):
        r"""
        The dimension of this wire.
        """
        return self._dim

    @dim.setter
    def dim(self, value: int):
        if self._dim:
            raise ValueError("Cannot change the dimension of wire with specified dimension.")
        self._dim = value

    @property
    def is_connected(self) -> bool:
        r"""
        Whether or not this wire is connected with another wire.
        """
        return self._is_connected

    @is_connected.setter
    def is_connected(self, value: bool):
        self._is_connected = value


@dataclass
class WireGroup:
    r"""A group of wires in a tensor network.

    Args:
        ket: A dictionary containing the wires on the ket side.
        bra: A dictionary containing the wires on the bra side.

    """

    ket: dict = field(default_factory=dict)
    bra: dict = field(default_factory=dict)


class Tensor(ABC):
    r"""An abstract class representing a tensor in a tensor network.

    In Mr Mustard, tensors are used to represent a state or a transformation on a given set
    of modes in the Fock representation. For example, a single-mode unitary matrix
    :math:`U=\sum_{i,j=1}^Nu_{i,j}|i\rangle\langle{j}|` acting on mode ``3`` in an
    N-dimensional Fock basis corresponds to the following ``Tensor`` object:

    .. code-block::
        class U(Tensor):
            def value(self, shape):
                # specify the value of the tensor
                pass

        U("my_unitary", [3], [3], [3], [3])

    Args:
        name: The name of this tensor.
        modes_in_ket: The input modes on the ket side.
        modes_out_ket: The output modes on the ket side.
        modes_in_bra: The input modes on the bra side.
        modes_out_bra: The output modes on the bra side.
    """

    _repr_markdown_ = None  # otherwise it takes over the repr due to mro

    def __init__(
        self,
        name: str,
        modes_in_ket: Optional[list[int]] = None,
        modes_out_ket: Optional[list[int]] = None,
        modes_in_bra: Optional[list[int]] = None,
        modes_out_bra: Optional[list[int]] = None,
    ) -> None:
        self._name = name
        self._update_modes(modes_in_ket, modes_out_ket, modes_in_bra, modes_out_bra)

    def _update_modes(
        self,
        modes_in_ket: Optional[list[int]] = None,
        modes_out_ket: Optional[list[int]] = None,
        modes_in_bra: Optional[list[int]] = None,
        modes_out_bra: Optional[list[int]] = None,
    ) -> None:
        r"""
        Updates the modes in this tensor by setting:

          * self._modes_in_ket, a list of input modes on the ket side
          * self._modes_out_ket, a list of output modes on the ket side
          * self._modes_in_bra, a list of input modes on the bra side
          * self._modes_out_bra, a list of output modes on the bra side
          * self.self._input, a WireGroup containing all the input modes
          * self.self._output, a WireGroup containing all the output modes

        It computes a new ``id`` for every wire.

        Raises:
            ValueError: if `modes_in_ket` and `modes_in_bra` are not equal, and neither
                of them is `None`.
            ValueError: if `modes_out_ket` and `modes_out_bra` are not equal, and neither
                of them is `None`.
        """
        msg = "modes on ket and bra sides must be equal, unless either of them is `None`."
        if modes_in_ket and modes_in_bra:
            if modes_in_ket != modes_in_bra:
                msg = f"Input {msg}"
                raise ValueError(msg)
        if modes_out_ket and modes_out_bra:
            if modes_out_ket != modes_out_bra:
                msg = f"Output {msg}"
                raise ValueError(msg)

        self._modes_in_ket = modes_in_ket if modes_in_ket else []
        self._modes_out_ket = modes_out_ket if modes_out_ket else []
        self._modes_in_bra = modes_in_bra if modes_in_bra else []
        self._modes_out_bra = modes_out_bra if modes_out_bra else []

        # initialize ket and bra wire dicts
        self._input = WireGroup()
        for mode in self._modes_in_ket:
            self._input.ket |= {mode: Wire(random_int(), mode, True, True)}
        for mode in self._modes_in_bra:
            self._input.bra |= {mode: Wire(random_int(), mode, True, False)}

        self._output = WireGroup()
        for mode in self._modes_out_ket:
            self._output.ket |= {mode: Wire(random_int(), mode, False, True)}
        for mode in self._modes_out_bra:
            self._output.bra |= {mode: Wire(random_int(), mode, False, False)}

    @property
    def adjoint(self) -> AdjointView:
        r"""The adjoint view of this Tensor (with new ``id``s). That is, ket <-> bra."""
        return AdjointView(self)

    @property
    def input(self):
        r"""
        A dictionary mapping the input modes to their respective wires.
        """
        return self._input

    @property
    def modes(self) -> list[int]:
        r"""
        For backward compatibility. Don't overuse.
        It returns a list of modes for this Tensor, unless it's ambiguous.
        """
        if self.modes_in == self.modes_out:  # transformation on same modes
            return list(self.modes_in)
        elif len(self.modes_in) == 0:  # state
            return list(self.modes_out)
        elif len(self.modes_out) == 0:  # measurement
            return list(self.modes_in)
        else:
            raise ValueError("modes are ambiguous for this Tensor.")

    @property
    def modes_in(self) -> List[int]:
        r"""
        The list of input modes that are used by this Tensor.

        If this tensor has no input modes on the bra side, or if the input modes are equal
        on both ket and bra sides, it returns the list of modes. Otherwise, it performs the
        ``set()`` operation before returning the list (and hence, the order may be unexpected).
        """
        if self._modes_in_ket:
            return self._modes_in_ket
        return self._modes_in_bra

    @property
    def modes_out(self) -> List[int]:
        r"""
        The list of output modes that are used by this Tensor.

        If this tensor has no output modes on the bra side, or if the output modes are equal
        on both ket and bra sides, it returns the list of modes. Otherwise, it performs the
        ``set()`` operation before returning the list (and hence, the order may be unexpected).
        """
        if self._modes_out_ket:
            return self._modes_out_ket
        return self._modes_out_bra

    @property
    def name(self) -> int:
        r"""
        The name of this tensor.
        """
        return self._name

    @property
    def output(self):
        r"""
        A dictionary mapping the output modes to their respective wires.
        """
        return self._output

    def unpack_shape(self, shape: Tuple[int]):
        r"""
        Unpack the given ``shape`` into the shapes of the input and output wires on ket and bra sides.

        Args:
            shape: A shape.

        Returns:
            shape_in_ket: The shape of the input wires on the ket side.
            shape_out_ket: The shape of the output wires on the ket side.
            shape_in_bra: The shape of the input wires on the bra side.
            shape_out_bra: The shape of the output wires on the bra side.
        """
        idx1 = 0
        idx2 = len(self._modes_in_ket)
        shape_in_ket = shape[idx1:idx2]

        idx1 = idx2
        idx2 += len(self._modes_out_ket)
        shape_out_ket = shape[idx1:idx2]

        idx1 = idx2
        idx2 += len(self._modes_in_bra)
        shape_in_bra = shape[idx1:idx2]

        idx1 = idx2
        idx2 += len(self._modes_out_bra)
        shape_out_bra = shape[idx1:idx2]

        return shape_in_ket, shape_out_ket, shape_in_bra, shape_out_bra

    @property
    def wires(self) -> List[Wire]:
        r"""
        The list of all wires in this tensor, sorted as ``[ket_in, ket_out, bra_in, bra_out]``.
        """
        return (
            list(self.input.ket.values())
            + list(self.output.ket.values())
            + list(self.input.bra.values())
            + list(self.output.bra.values())
        )

    @abstractmethod
    def value(self, shape: Tuple[int]):
        r"""The value of this tensor.

        Args:
            shape: the shape of this tensor

        Returns:
            ComplexTensor: the unitary matrix in Fock representation
        """

    def change_modes(
        self,
        modes_in_ket: Optional[list[int]] = None,
        modes_out_ket: Optional[list[int]] = None,
        modes_in_bra: Optional[list[int]] = None,
        modes_out_bra: Optional[list[int]] = None,
    ) -> None:
        r"""
        Changes the modes in this tensor.

        Args:
            name: The name of this tensor.
            modes_in_ket: The input modes on the ket side.
            modes_out_ket: The output modes on the ket side.
            modes_in_bra: The input modes on the bra side.
            modes_out_bra: The output modes on the bra side.

        Raises:
            ValueError: if one or more wires in this tensor are already connected.
        """
        for wire in self.wires:
            if wire.is_connected:
                msg = (
                    "Cannot change nodes in a tensor when some of its wires are already connected."
                )
                raise ValueError(msg)
        self._update_modes(modes_in_ket, modes_out_ket, modes_in_bra, modes_out_bra)

    def shape(self, default_dim: Optional[int] = None, out_in=False):
        r"""
        Returns the shape of the underlying tensor, as inferred from the dimensions of the individual
        wires.

        If ``out_in`` is ``False``, the shape returned is in the order ``(in_ket, in_bra, out_ket, out_bra)``.
        Otherwise, it is in the order ``(out_ket, out_bra, in_ket, in_bra)``.

        Args:
            default_dim: The default dimension of wires with unspecified dimension.
            out_in: Whether to return output shapes followed by input shapes or viceversa.
        """

        def _sort_shapes(*args):
            for arg in args:
                if arg:
                    yield arg

        shape_in_ket = [w.dim if w.dim else default_dim for w in self.input.ket.values()]
        shape_out_ket = [w.dim if w.dim else default_dim for w in self.output.ket.values()]
        shape_in_bra = [w.dim if w.dim else default_dim for w in self.input.bra.values()]
        shape_out_bra = [w.dim if w.dim else default_dim for w in self.output.bra.values()]

        if out_in:
            ret = _sort_shapes(shape_out_ket, shape_out_bra, shape_in_ket, shape_in_bra)
        ret = _sort_shapes(shape_in_ket, shape_in_bra, shape_out_ket, shape_out_bra)

        # pylint: disable=consider-using-generator
        return tuple([item for sublist in ret for item in sublist])


class AdjointView(Tensor):
    r"""
    Adjoint view of a tensor. It swaps the ket and bra wires of a Tensor.
    """

    def __init__(self, tensor):
        self._original = tensor
        super().__init__(
            name=self._original.name,
            modes_in_ket=self._original.input.bra.keys(),
            modes_out_ket=self._original.output.bra.keys(),
            modes_in_bra=self._original.input.ket.keys(),
            modes_out_bra=self._original.output.ket.keys(),
        )

    def value(self, shape: Tuple[int]):
        r"""The value of this tensor.

        Args:
            shape: the shape of the adjoint tensor.

        Returns:
            ComplexTensor: the unitary matrix in Fock representation
        """
        # converting the given shape into a shape for the original tensor
        shape_in_ket, shape_out_ket, shape_in_bra, shape_out_bra = self._original.unpack_shape(
            shape
        )
        shape_ret = shape_in_bra + shape_out_bra + shape_in_ket + shape_out_ket

        ret = math.conj(math.astensor(self._original.value(shape_ret)))
        return ret


class DualView(Tensor):
    r"""
    Dual view of a tensor. It swaps the input and output wires of a Tensor.
    """

    def __init__(self, tensor):
        self._original = tensor
        super().__init__(
            name=self._original.name,
            modes_in_ket=self._original.output.ket.keys(),
            modes_out_ket=self._original.input.ket.keys(),
            modes_in_bra=self._original.output.bra.keys(),
            modes_out_bra=self._original.input.bra.keys(),
        )

    def value(self, shape: Tuple[int]):
        r"""The value of this tensor.

        Args:
            shape: the shape of the dual tensor.

        Returns:
            ComplexTensor: the unitary matrix in Fock representation.
        """
        # converting the given shape into a shape for the original tensor
        shape_in_ket, shape_out_ket, shape_in_bra, shape_out_bra = self.unpack_shape(shape)
        shape_ret = shape_out_ket + shape_in_ket + shape_out_bra, shape_in_bra

        return math.conj(self._original.value(shape_ret))



# Copyright 2022 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This module contains the implementation of distributed training utilities for parallelized
optimization of MrMustard circuits/devices through the function :meth:`map_trainer`.

This module requires extra dependencies, to install:

.. code-block:: bash

    git clone https://github.com/XanaduAI/MrMustard
    cd MrMustard
    pip install -e .[ray]


User-provided Wrapper Functions
===============================
To distribute your optimization workflow, two user-defined functions are needed for wrapping up user logic:

* A `device_factory` that wraps around the logic for making your circuits/states to be optimized; it is expected to return a single, or list of, :class:`Circuit`(s).
* A `cost_fn` that takes the circuits made and additional keyword arguments and returns a backprop-able scalar cost.

Separating the circuit-making logic from the cost calculation logic has the benefit of returning the optimized circuit in the result dict for further inspection. One can also pass extra `metric_fns` to directly extract info from the circuit.


Examples:
=========

.. code-block::

    from mrmustard.lab import Vacuum, Dgate, Ggate, Gaussian
    from mrmustard.physics import fidelity
    from mrmustard.training.trainer import map_trainer

    def make_circ(x=0.):
        return Ggate(num_modes=1, symplectic_trainable=True) >> Dgate(x=x, x_trainable=True, y_trainable=True)

    def cost_fn(circ=make_circ(0.1), y_targ=0.):
        target = Gaussian(1) >> Dgate(-1.5, y_targ)
        s = Vacuum(1) >> circ
        return -fidelity(s, target)

    # Use case 0: Calculate the cost of a randomly initialized circuit 5 times without optimizing it.
    results_0 = map_trainer(
        cost_fn=cost_fn,
        tasks=5,
    )

    # Use case 1: Run circuit optimization 5 times on randomly initialized circuits.
    results_1 = map_trainer(
        cost_fn=cost_fn,
        device_factory=make_circ,
        tasks=5,
        max_steps=50,
        symplectic_lr=0.05,
    )

    # Use case 2: Run 2 sets of circuit optimization with custom parameters passed as list.
    results_2 = map_trainer(
        cost_fn=cost_fn,
        device_factory=make_circ,
        tasks=[
            {'x': 0.1, 'euclidean_lr': 0.005, 'max_steps': 50},
            {'x': -0.7, 'euclidean_lr': 0.1, 'max_steps': 2},
        ],
        y_targ=0.35,
        symplectic_lr=0.05,
        AUTOCUTOFF_MAX_CUTOFF=7,
    )

    # Use case 3: Run 2 sets of circuit optimization with custom parameters passed as dict with extra metric functions for evaluating the final optimized circuit.
    results_3 = map_trainer(
    cost_fn=cost_fn,
    device_factory=make_circ,
    tasks={
        'my-job': {'x': 0.1, 'euclidean_lr': 0.005, 'max_steps': 50},
        'my-other-job': {'x': -0.7, 'euclidean_lr': 0.1, 'max_steps': 2},
    },
    y_targ=0.35,
    symplectic_lr=0.05,
    metric_fns={
        'is_gaussian': lambda c: c.is_gaussian,
        'foo': lambda _: 17.
    },
)


"""

import warnings
from functools import partial
from inspect import Parameter, signature
from typing import Mapping, Sequence

import numpy as np
from rich.progress import track

from mrmustard import settings
from .optimizer import Optimizer


def _apply_partial_cost(device, cost_fn, **kwargs):
    """Helper partial cost fn maker."""
    if isinstance(device, Sequence):
        cost_fn, kwargs = partial_pop(cost_fn, *device, **kwargs)
        optimized = device
    elif isinstance(device, Mapping):
        cost_fn, kwargs = partial_pop(cost_fn, **device, **kwargs)
        optimized = list(device.values())
    return cost_fn, kwargs, optimized


def train_device(
    cost_fn,
    device_factory=None,
    metric_fns=None,
    return_kwargs=True,
    skip_opt=False,
    tag=None,
    **kwargs,
):
    """A general and flexible training loop for circuit optimizations with configurations adjustable through kwargs.

    Args:
        cost_fn (callable): The optimized cost function to be distributed. It's expected to accept the
            output of `device_factory` as *args as well as user-defined **kwargs, and returns a scalar cost.
            Its user-defined **kwargs will be passed from this function's **kwargs which must include all its
            required arguments.
        device_factory (callable): Function that (partially) takes `kwargs` and returns a device, or
            list/dict of devices. If None, `cost_fn` will be assumed to take no positional argument (for
            example, when device-making is contained in `cost_fn`). Defaults to None.
        metric_fns (Union[Sequence[callable], Mapping[callable], callable]): Optional collection of functions that takes the
            output of `device_factory` after optimization and returns arbitrary evaluation/information.
        return_kwargs (bool): Whether to include input config `kwargs` in the output dict. Defualts to True.
        skip_opt (bool): Whether to skip the optimization and directly calculate cost.
        tag (str): Optional label of the training task associated with the `kwargs` to be included in the output dict.
        kwargs:
            Dict containing all arguments to any of the functions below:
                - `cost_fn`: exluding the output of `device_factory`.
                - `device_factory`: e.g. `x`, `r`, `theta`, etc.
                - `Optimizer`: e.g. `euclidean_lr`.
                - `Optimizer.minimize`: excluding `cost_fn` and `by_optimizing`, e.g. `max_steps`.

    Returns:
        dict: A result dict summarizing the optimized circuit, cost, metrics and/or input configs.

    """

    setting_updates, kwargs = update_pop(settings, **kwargs)

    input_kwargs = kwargs.copy() if return_kwargs else {}

    device, kwargs = (
        curry_pop(device_factory, **kwargs) if callable(device_factory) else ([], kwargs)
    )
    device = [device] if not isinstance(device, (Sequence, Mapping)) else device

    cost_fn, kwargs, optimized = _apply_partial_cost(device, cost_fn, **kwargs)

    opt = None
    if optimized and not skip_opt:
        opt, kwargs = curry_pop(Optimizer, **kwargs)
        _, kwargs = curry_pop(
            opt.minimize, **{"cost_fn": cost_fn, "by_optimizing": optimized}, **kwargs
        )

    if kwargs:
        warnings.warn(f"Unused kwargs: {kwargs}")

    final_cost = cost_fn()

    results = {
        "cost": np.array(final_cost).item(),
        "device": device,
        "optimizer": opt,
    }

    if callable(metric_fns):
        results["metrics"] = metric_fns(*device)
    elif isinstance(metric_fns, Sequence):
        results["metrics"] = [f(*device) for f in metric_fns if callable(f)]
    elif isinstance(metric_fns, Mapping):
        results = {
            **results,
            **{k: f(*device) for k, f in metric_fns.items() if callable(f)},
        }

    return {
        **({"tag": tag} if tag is not None else {}),
        **results,
        **input_kwargs,
        **setting_updates,
    }


def _iter_futures(futures):
    """Make ray futures iterable for easy passing to a progress bar.
    Hacky: https://github.com/ray-project/ray/issues/5554
    """
    import ray  # pylint: disable=import-outside-toplevel

    while futures:
        done, futures = ray.wait(futures)
        yield ray.get(done[0])


def map_trainer(trainer=train_device, tasks=1, pbar=True, unblock=False, num_cpus=None, **kwargs):
    """Maps multiple training tasks across multiple workers using `ray`.

    In practice, the most common use case is to ignore the keywords `trainer` (as it defaults to
    :meth:`train_device`), `pbar`, `unblock`, etc. and just concentrate on `tasks` and `**kwargs`
    which passes arguments to the wrapper functions that contain the task execution logic, as well
    as the :class:`Optimizer` and its :meth:`Optimizer.minimize`.

    For example, with the default `trainer` :meth:`train_device`, two user-defined functions are used for wrapping up user logic:

    * A `device_factory` (optional) that wraps around the logic for making circuits/states to be optimized; it is expected to return a single, or list of, :class:`Circuit`(s).

    * A `cost_fn` (required) that takes the circuits made and additional keyword arguments and returns a backprop-able scalar cost.

    Refer to the `kwargs` section below for more available options.

    Args:
        trainer (callable): The function containing the training loop to be distributed, whose
            fixed arguments are to be passed by `**kwargs` and task-specific arguments iterated
            through `tasks`. Provide only when custom evaluation/training logic is needed.
            Defaults to :meth:`train_device`.
        tasks (Union[int, Sequence, Mapping]): Number of repeats or collection of task-specific training
            config arguments feeding into :meth:`train_device`.
            Refer to `kwargs` below for the available options.
            Defaults to 1 which runs `trainer` exactly once.
        pbar (bool): Whether to show a progress bar, available only in blocking mode (i.e. `unblock==False`). Defaults to True.
        unblock (bool): Whether to unblock the process and returns a getter function returning the available results.
            Defaults to False.
        num_cpus (int): Number of cpu workers to initialize ray. Defaults to the number of virtual cores.
        kwargs: Additional arguments containing fixed training config kwargs feeding into `trainer`.
            For the default `trainer` :meth:`train_device`, available options are:
                - cost_fn (callable):
                    The optimized cost function to be distributed. It's expected to accept the
                    output of `device_factory` as *args as well as user-defined **kwargs, and returns a scalar cost.
                    Its user-defined **kwargs will be passed from this function's **kwargs which must include all its
                    required arguments.
                - device_factory (callable):
                    Function that (partially) takes `kwargs` and returns a device, or
                    list/dict of devices. If None, `cost_fn` will be assumed to take no positional argument (for
                    example, when device-making is contained in `cost_fn`). Defaults to None.
                - metric_fns (Union[Sequence[callable], Mapping[callable], callable]):
                    Optional collection of functions that takes the
                    output of `device_factory` after optimization and returns arbitrary evaluation/information.
                - return_kwargs (bool):
                    Whether to include input config `kwargs` in the output dict. Defualts to True.
                - skip_opt (bool):
                    Whether to skip the optimization and directly calculate cost.
                - tag (str):
                    Optional label of the training task associated with the `kwargs` to be included in the output dict.
                - any kwargs to `cost_fn`: exluding the output of `device_factory`.
                - any kwargs to `device_factory`: e.g. `x`, `r`, `theta`, etc.
                - any kwargs to `Optimizer`: e.g. `euclidean_lr`.
                - any kwargs to `Optimizer.minimize`: excluding `cost_fn` and `by_optimizing`, e.g. `max_steps`.

    Returns
        Union[List, Dict]: The collection of results from each training task. Returns
            - a list if `tasks` is provided as an int or a list; or
            - a dict with the same keys if `tasks` is provided as a dict.


    Examples:
    =========

    .. code-block::

        from mrmustard.lab import Vacuum, Dgate, Ggate, Gaussian
        from mrmustard.physics import fidelity
        from mrmustard.training.trainer import map_trainer

        def make_circ(x=0.):
            return Ggate(num_modes=1, symplectic_trainable=True) >> Dgate(x=x, x_trainable=True, y_trainable=True)

        def cost_fn(circ=make_circ(0.1), y_targ=0.):
            target = Gaussian(1) >> Dgate(-1.5, y_targ)
            s = Vacuum(1) >> circ
            return -fidelity(s, target)

        # Use case 0: Calculate the cost of a randomly initialized circuit 5 times without optimizing it.
        results_0 = map_trainer(
            cost_fn=cost_fn,
            tasks=5,
        )

        # Use case 1: Run circuit optimization 5 times on randomly initialized circuits.
        results_1 = map_trainer(
            cost_fn=cost_fn,
            device_factory=make_circ,
            tasks=5,
            max_steps=50,
            symplectic_lr=0.05,
        )

        # Use case 2: Run 2 sets of circuit optimization with custom parameters passed as list.
        results_2 = map_trainer(
            cost_fn=cost_fn,
            device_factory=make_circ,
            tasks=[
                {'x': 0.1, 'euclidean_lr': 0.005, 'max_steps': 50},
                {'x': -0.7, 'euclidean_lr': 0.1, 'max_steps': 2},
            ],
            y_targ=0.35,
            symplectic_lr=0.05,
            AUTOCUTOFF_MAX_CUTOFF=7,
        )

        # Use case 3: Run 2 sets of circuit optimization with custom parameters passed as dict with extra metric functions for evaluating the final optimized circuit.
        results_3 = map_trainer(
        cost_fn=cost_fn,
        device_factory=make_circ,
        tasks={
            'my-job': {'x': 0.1, 'euclidean_lr': 0.005, 'max_steps': 50},
            'my-other-job': {'x': -0.7, 'euclidean_lr': 0.1, 'max_steps': 2},
        },
        y_targ=0.35,
        symplectic_lr=0.05,
        metric_fns={
            'is_gaussian': lambda c: c.is_gaussian,
            'foo': lambda _: 17.
        },
    )
    """
    try:
        import ray  # pylint: disable=import-outside-toplevel
    except ImportError as e:
        raise ImportError(
            "Failed to import `ray` which is an extra dependency. Please install with `pip install -e .[ray]`."
        ) from e

    if not ray.is_initialized():  # pragma: no cover
        ray.init(num_cpus=num_cpus)

    return_dict = False
    if isinstance(tasks, int):
        tasks = [{} for _ in range(tasks)]
    elif isinstance(tasks, Mapping):
        return_dict = True
        tasks = [{"tag": tag, **task} for tag, task in tasks.items()]

    remote_trainer, kwargs = partial_pop(
        ray.remote(trainer).remote,
        **kwargs,
    )

    if isinstance(tasks, Sequence):
        promises = [
            curry_pop(
                remote_trainer,
                **task_kwargs,
            )[0]
            for task_kwargs in tasks
            if isinstance(task_kwargs, Mapping)
        ]
    else:
        raise ValueError(
            f"`tasks` is expected to be of type int, list, or dict. got {type(tasks)}: {tasks}"
        )

    if not unblock:
        # blocks and wait till all tasks complete to return the end results.
        if pbar:
            results = list(
                track(
                    _iter_futures(promises),
                    description=f"{len(promises)} tasks running...",
                    total=len(promises),
                )
            )
        else:
            results = ray.get(promises)

        if return_dict:
            return {r["tag"]: r for r in results}
        else:
            return results

    else:
        # does not block and returns a getter function that returns the available results so far.
        def get_avail_results():
            results, running_tasks = ray.wait(  # pylint: disable=unused-variable
                promises, num_returns=len(promises)
            )
            if return_dict:
                return {r["tag"]: r for r in ray.get(results)}
            else:
                return ray.get(results)

        return get_avail_results


def kwargs_of(fn):
    """Gets the kwarg signature of a callable."""
    params = signature(fn).parameters
    kwarg_kinds = [Parameter.POSITIONAL_OR_KEYWORD, Parameter.KEYWORD_ONLY]

    keywords = [k for k, p in params.items() if p.kind in kwarg_kinds]
    has_var_keyword = any(p.kind is Parameter.VAR_KEYWORD for p in params.values())

    return keywords, has_var_keyword


def partial_pop(fn, *args, **kwargs):
    """Partially applies known kwargs to fn and returns the rest."""
    keywords, has_var_keyword = kwargs_of(fn)
    known_kwargs = {k: kwargs.pop(k) for k in set(kwargs).intersection(keywords)}
    partial_fn = partial(fn, *args, **known_kwargs, **(kwargs if has_var_keyword else {}))
    return partial_fn, kwargs


def curry_pop(fn, *args, **kwargs):
    """A poor man's reader monad bind function."""
    partial_fn, kwargs = partial_pop(fn, *args, **kwargs)
    return partial_fn(), kwargs


def update_pop(obj, **kwargs):
    """Updates an object/dict while popping keys out and returns the updated dict and remaining kwargs."""
    updated = {}
    if isinstance(obj, Mapping):
        for k in set(kwargs).intersection(obj):
            obj[k] = kwargs.pop(k)
            updated[k] = obj[k]
    else:
        for k in set(kwargs).intersection(dir(obj)):
            setattr(obj, k, kwargs.pop(k))
            updated[k] = getattr(obj, k)
    return updated, kwargs



# Copyright 2021 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This module contains the implementation of the :class:`State` class."""

from __future__ import annotations

import warnings
from typing import (
    TYPE_CHECKING,
    Iterable,
    List,
    Optional,
    Sequence,
    Tuple,
    Union,
)

import matplotlib.pyplot as plt
import numpy as np
from matplotlib import cm

from mrmustard import math, settings
from mrmustard.math.parameters import Constant, Variable
from mrmustard.physics import bargmann, fock, gaussian
from mrmustard.physics.wigner import wigner_discretized
from mrmustard.utils.typing import (
    ComplexMatrix,
    ComplexTensor,
    ComplexVector,
    RealMatrix,
    RealTensor,
    RealVector,
)

if TYPE_CHECKING:
    from .transformation import Transformation


# pylint: disable=too-many-instance-attributes
class State:  # pylint: disable=too-many-public-methods
    r"""Base class for quantum states."""

    def __init__(
        self,
        cov: RealMatrix = None,
        means: RealVector = None,
        eigenvalues: RealVector = None,
        symplectic: RealMatrix = None,
        ket: ComplexTensor = None,
        dm: ComplexTensor = None,
        modes: Sequence[int] = None,
        cutoffs: Sequence[int] = None,
        _norm: float = 1.0,
    ):
        r"""Initializes the state.

        Supply either:
            * a covariance matrix and means vector
            * an eigenvalues array and symplectic matrix
            * a fock representation (ket or dm)

        Args:
            cov (Matrix): the covariance matrix
            means (Vector): the means vector
            eigenvalues (Tensor): the eigenvalues of the covariance matrix
            symplectic (Matrix): the symplectic matrix mapping the thermal state with given eigenvalues to this state
            fock (Tensor): the Fock representation
            modes (optional, Sequence[int]): the modes in which the state is defined
            cutoffs (Sequence[int], default=None): set to force the cutoff dimensions of the state
            _norm (float, default=1.0): the norm of the state. Warning: only set if you know what you are doing.

        """
        self._purity = None
        self._fock_probabilities = None
        self._cutoffs = cutoffs
        self._cov = cov
        self._means = means
        self._eigenvalues = eigenvalues
        self._symplectic = symplectic
        self._ket = ket
        self._dm = dm
        self._norm = _norm
        if cov is not None and means is not None:
            self.is_gaussian = True
            self.is_hilbert_vector = np.allclose(gaussian.purity(self.cov), 1.0, atol=1e-6)
            self.num_modes = cov.shape[-1] // 2
        elif eigenvalues is not None and symplectic is not None:
            self.is_gaussian = True
            self.is_hilbert_vector = np.allclose(eigenvalues, 2.0 / settings.HBAR)
            self.num_modes = symplectic.shape[-1] // 2
        elif ket is not None or dm is not None:
            self.is_gaussian = False
            self.is_hilbert_vector = ket is not None
            self.num_modes = len(ket.shape) if ket is not None else len(dm.shape) // 2
            self._purity = 1.0 if ket is not None else None
        else:
            raise ValueError(
                "State must be initialized with either a covariance matrix and means vector, an eigenvalues array and symplectic matrix, or a fock representation"
            )
        self._modes = modes
        if modes is not None:
            assert (
                len(modes) == self.num_modes
            ), f"Number of modes supplied ({len(modes)}) must match the representation dimension {self.num_modes}"

    def _add_parameter(self, parameter: Union[Constant, Variable]):
        r"""
        Adds a parameter to a state.

        Args:
            parameter: The parameter to add.
        """
        if not getattr(self, "_parameter_set", None):
            msg = "Cannot add a parameter to a state with no parameter set."
            raise ValueError(msg)
        self.parameter_set.add_parameter(parameter)
        self.__dict__[parameter.name] = parameter

    @property
    def parameter_set(self):
        r"""
        The set of parameters for this state.
        """
        return getattr(self, "_parameter_set", None)

    @property
    def modes(self):
        r"""Returns the modes of the state."""
        if self._modes is None:
            return list(range(self.num_modes))
        return self._modes

    def indices(self, modes) -> Union[Tuple[int], int]:
        r"""Returns the indices of the given modes.

        Args:
            modes (Sequence[int] or int): the modes or mode

        Returns:
            Tuple[int] or int: a tuple of indices of the given modes or the single index of a single mode
        """
        if isinstance(modes, int):
            return self.modes.index(modes)
        return tuple(self.modes.index(m) for m in modes)

    @property
    def purity(self) -> float:
        """Returns the purity of the state."""
        if self._purity is None:
            if self.is_gaussian:
                self._purity = gaussian.purity(self.cov)
            else:
                self._purity = fock.purity(self._dm)
        return self._purity

    @property
    def is_mixed(self):
        r"""Returns whether the state is mixed."""
        return not self.is_pure

    @property
    def is_pure(self):
        r"""Returns ``True`` if the state is pure and ``False`` otherwise."""
        return np.isclose(self.purity, 1.0, atol=1e-6)

    @property
    def means(self) -> Optional[RealVector]:
        r"""Returns the means vector of the state."""
        return self._means

    @property
    def cov(self) -> Optional[RealMatrix]:
        r"""Returns the covariance matrix of the state."""
        return self._cov

    @property
    def number_stdev(self) -> RealVector:
        r"""Returns the square root of the photon number variances (standard deviation) in each mode."""
        if self.is_gaussian:
            return math.sqrt(math.diag_part(self.number_cov))

        return math.sqrt(
            fock.number_variances(self.fock, is_dm=len(self.fock.shape) == self.num_modes * 2)
        )

    @property
    def cutoffs(self) -> List[int]:
        r"""Returns the Hilbert space dimension of each mode."""
        if self._cutoffs is None:
            if self._ket is None and self._dm is None:
                self._cutoffs = fock.autocutoffs(
                    self.cov, self.means, settings.AUTOCUTOFF_PROBABILITY
                )
            else:
                self._cutoffs = [
                    int(c)
                    for c in (
                        self._ket.shape
                        if self._ket is not None
                        else self._dm.shape[: self.num_modes]
                    )
                ]
        return self._cutoffs

    @property
    def shape(self) -> List[int]:
        r"""Returns the shape of the state, accounting for ket/dm representation.

        If the state is in Gaussian representation, the shape is inferred from
        the first two moments of the number operator.
        """
        # NOTE: if we initialize State(dm=pure_dm), self.fock returns the dm, which does not have shape self.cutoffs
        return self.cutoffs if self.is_hilbert_vector else self.cutoffs + self.cutoffs

    @property
    def fock(self) -> ComplexTensor:
        r"""Returns the Fock representation of the state."""
        if self._dm is None and self._ket is None:
            _fock = fock.wigner_to_fock_state(
                self.cov,
                self.means,
                shape=self.shape,
                return_dm=not self.is_hilbert_vector,
            )
            if self.is_mixed:
                self._dm = _fock
                self._ket = None
            else:
                self._ket = _fock
                self._dm = None
        return self._ket if self._ket is not None else self._dm

    @property
    def number_means(self) -> RealVector:
        r"""Returns the mean photon number for each mode."""
        if self.is_gaussian:
            return gaussian.number_means(self.cov, self.means)

        return fock.number_means(tensor=self.fock, is_dm=self.is_mixed)

    @property
    def number_cov(self) -> RealMatrix:
        r"""Returns the complete photon number covariance matrix."""
        if not self.is_gaussian:
            raise NotImplementedError("number_cov not yet implemented for non-gaussian states")

        return gaussian.number_cov(self.cov, self.means)

    @property
    def norm(self) -> float:
        r"""Returns the norm of the state."""
        if self.is_gaussian:
            return self._norm
        return fock.norm(self.fock, not self.is_hilbert_vector)

    @property
    def probability(self) -> float:
        r"""Returns the probability of the state."""
        norm = self.norm
        if self.is_pure and self._ket is not None:
            return norm**2
        return norm

    def ket(
        self,
        cutoffs: List[int] = None,
        max_prob: float = 1.0,
        max_photons: int = None,
    ) -> Optional[ComplexTensor]:
        r"""Returns the ket of the state in Fock representation or ``None`` if the state is mixed.

        Args:
            cutoffs List[int or None]: The cutoff dimensions for each mode. If a mode cutoff is
                ``None``, it's guessed automatically.
            max_prob (float): The maximum probability of the state. Defaults to 1.0.
                (used to stop the calculation of the amplitudes early)
            max_photons (int): The maximum number of photons in the state, summing over all modes
                (used to stop the calculation of the amplitudes early)

        Returns:
            Tensor: the ket
        """
        if self.is_mixed:
            return None

        if cutoffs is None:
            cutoffs = self.cutoffs
        else:
            cutoffs = [c if c is not None else self.cutoffs[i] for i, c in enumerate(cutoffs)]

        # TODO: shouldn't we check if trainable instead? that's when we want to recompute fock
        if self.is_gaussian:
            self._ket = fock.wigner_to_fock_state(
                self.cov,
                self.means,
                shape=cutoffs,
                return_dm=False,
                max_prob=max_prob,
                max_photons=max_photons,
            )
        else:  # only fock representation is available
            if self._ket is None:
                # if state is pure and has a density matrix, calculate the ket
                if self.is_pure:
                    self._ket = fock.dm_to_ket(self._dm)
            current_cutoffs = [int(s) for s in self._ket.shape]
            if cutoffs != current_cutoffs:
                paddings = [(0, max(0, new - old)) for new, old in zip(cutoffs, current_cutoffs)]
                if any(p != (0, 0) for p in paddings):
                    padded = fock.math.pad(self._ket, paddings, mode="constant")
                else:
                    padded = self._ket
                return padded[tuple(slice(s) for s in cutoffs)]
        return self._ket[tuple(slice(s) for s in cutoffs)]

    def dm(self, cutoffs: Optional[List[int]] = None) -> ComplexTensor:
        r"""Returns the density matrix of the state in Fock representation.

        Args:
            cutoffs List[int]: The cutoff dimensions for each mode. If a mode cutoff is ``None``,
                it's automatically computed.

        Returns:
            Tensor: the density matrix
        """
        if cutoffs is None:
            cutoffs = self.cutoffs
        else:
            cutoffs = [c if c is not None else self.cutoffs[i] for i, c in enumerate(cutoffs)]
        if self.is_pure:
            ket = self.ket(cutoffs=cutoffs)
            if ket is not None:
                return fock.ket_to_dm(ket)
        else:
            if self.is_gaussian:
                self._dm = fock.wigner_to_fock_state(
                    self.cov, self.means, shape=cutoffs + cutoffs, return_dm=True
                )
            elif cutoffs != (current_cutoffs := list(self._dm.shape[: self.num_modes])):
                paddings = [(0, max(0, new - old)) for new, old in zip(cutoffs, current_cutoffs)]
                if any(p != (0, 0) for p in paddings):
                    padded = fock.math.pad(self._dm, paddings + paddings, mode="constant")
                else:
                    padded = self._dm
                return padded[tuple(slice(s) for s in cutoffs + cutoffs)]
        return self._dm[tuple(slice(s) for s in cutoffs + cutoffs)]

    def fock_probabilities(self, cutoffs: Sequence[int]) -> RealTensor:
        r"""Returns the probabilities in Fock representation.

        If the state is pure, they are the absolute value squared of the ket amplitudes.
        If the state is mixed they are the multi-dimensional diagonals of the density matrix.

        Args:
            cutoffs List[int]: the cutoff dimensions for each mode

        Returns:
            Tensor: the probabilities
        """
        if self._fock_probabilities is None:
            if self.is_mixed:
                dm = self.dm(cutoffs=cutoffs)
                self._fock_probabilities = fock.dm_to_probs(dm)
            else:
                ket = self.ket(cutoffs=cutoffs)
                self._fock_probabilities = fock.ket_to_probs(ket)
        return self._fock_probabilities

    def primal(self, other: Union[State, Transformation]) -> State:
        r"""Returns the post-measurement state after ``other`` is projected onto ``self``.

        ``other << self`` is other projected onto ``self``.

        If ``other`` is a ``Transformation``, it returns the dual of the transformation applied to
        ``self``: ``other << self`` is like ``self >> other^dual``.

        Note that the returned state is not normalized. To normalize a state you can use
        ``mrmustard.physics.normalize``.
        """
        # import pdb

        # pdb.set_trace()
        if isinstance(other, State):
            return self._project_onto_state(other)
        try:
            return other.dual(self)
        except AttributeError as e:
            raise TypeError(
                f"Cannot apply {other.__class__.__qualname__} to {self.__class__.__qualname__}"
            ) from e

    def _project_onto_state(self, other: State) -> Union[State, float]:
        """If states are gaussian use generaldyne measurement, else use
        the states' Fock representation."""

        # if both states are gaussian
        if self.is_gaussian and other.is_gaussian:
            return self._project_onto_gaussian(other)

        # either self or other is not gaussian
        return self._project_onto_fock(other)

    def _project_onto_fock(self, other: State) -> Union[State, float]:
        """Returns the post-measurement state of the projection between two non-Gaussian
        states on the remaining modes or the probability of the result. When doing homodyne sampling,
        returns the post-measurement state or the measument outcome if no modes remain.

        Args:
            other (State): state being projected onto self

        Returns:
            State or float: returns the conditional state on the remaining modes
                or the probability.
        """
        remaining_modes = list(set(other.modes) - set(self.modes))

        out_fock = self._contract_with_other(other)
        if len(remaining_modes) > 0:
            return (
                State(dm=out_fock, modes=remaining_modes)
                if other.is_mixed or self.is_mixed
                else State(ket=out_fock, modes=remaining_modes)
            )

        # return the probability (norm) of the state when there are no modes left
        return (
            fock.math.abs(out_fock) ** 2
            if other.is_pure and self.is_pure
            else fock.math.abs(out_fock)
        )

    def _contract_with_other(self, other):
        other_cutoffs = [
            None if m not in self.modes else other.cutoffs[other.indices(m)] for m in other.modes
        ]
        if hasattr(self, "_preferred_projection"):
            out_fock = self._preferred_projection(other, other.indices(self.modes))
        else:
            # matching other's cutoffs
            self_cutoffs = [other.cutoffs[other.indices(m)] for m in self.modes]
            out_fock = fock.contract_states(
                stateA=other.ket(other_cutoffs) if other.is_pure else other.dm(other_cutoffs),
                stateB=self.ket(self_cutoffs) if self.is_pure else self.dm(self_cutoffs),
                a_is_dm=other.is_mixed,
                b_is_dm=self.is_mixed,
                modes=other.indices(self.modes),
                normalize=self._normalize if hasattr(self, "_normalize") else False,
            )

        return out_fock

    def _project_onto_gaussian(self, other: State) -> Union[State, float]:
        """Returns the result of a generaldyne measurement given that states ``self`` and
        ``other`` are gaussian.

        Args:
            other (State): gaussian state being projected onto self

        Returns:
            State or float: returns the output conditional state on the remaining modes
                or the probability.
        """
        # here `self` is the measurement device state and `other` is the incoming state
        # being projected onto the measurement state
        remaining_modes = list(set(other.modes) - set(self.modes))

        _, probability, new_cov, new_means = gaussian.general_dyne(
            other.cov,
            other.means,
            self.cov,
            self.means,
            self.modes,
        )

        if len(remaining_modes) > 0:
            return State(
                means=new_means,
                cov=new_cov,
                modes=remaining_modes,
                _norm=probability if not getattr(self, "_normalize", False) else 1.0,
            )

        return probability

    def __iter__(self) -> Iterable[State]:
        """Iterates over the modes and their corresponding tensors."""
        return (self.get_modes(i) for i in range(self.num_modes))

    def __and__(self, other: State) -> State:
        r"""Concatenates two states."""
        if not self.is_gaussian or not other.is_gaussian:  # convert all to fock now
            # TODO: would be more efficient if we could keep pure states as kets
            if self.is_mixed or other.is_mixed:
                self_fock = self.dm()
                other_fock = other.dm()
                dm = fock.math.tensordot(self_fock, other_fock, [[], []])
                # e.g. self has shape [1,3,1,3] and other has shape [2,2]
                # we want self & other to have shape [1,3,2,1,3,2]
                # before transposing shape is [1,3,1,3]+[2,2]
                self_idx = list(range(len(self_fock.shape)))
                other_idx = list(range(len(self_idx), len(self_idx) + len(other_fock.shape)))
                return State(
                    dm=math.transpose(
                        dm,
                        self_idx[: len(self_idx) // 2]
                        + other_idx[: len(other_idx) // 2]
                        + self_idx[len(self_idx) // 2 :]
                        + other_idx[len(other_idx) // 2 :],
                    ),
                    modes=self.modes + [m + max(self.modes) + 1 for m in other.modes],
                )
            # else, all states are pure
            self_fock = self.ket()
            other_fock = other.ket()
            return State(
                ket=fock.math.tensordot(self_fock, other_fock, [[], []]),
                modes=self.modes + [m + max(self.modes) + 1 for m in other.modes],
            )
        cov = gaussian.join_covs([self.cov, other.cov])
        means = gaussian.join_means([self.means, other.means])
        return State(
            cov=cov,
            means=means,
            modes=self.modes + [m + self.num_modes for m in other.modes],
        )

    def __getitem__(self, item) -> State:
        "setting the modes of a state (same API of `Transformation`)"
        if isinstance(item, int):
            item = [item]
        elif isinstance(item, Iterable):
            item = list(item)
        else:
            raise TypeError("item must be int or iterable")
        if len(item) != self.num_modes:
            raise ValueError(
                f"there are {self.num_modes} modes (item has {len(item)} elements, perhaps you're looking for .get_modes()?)"
            )
        self._modes = item
        return self

    def bargmann(self, numpy=False) -> Optional[tuple[ComplexMatrix, ComplexVector, complex]]:
        r"""Returns the Bargmann representation of the state.
        If numpy=True, returns the numpy arrays instead of the backend arrays.
        """
        if self.is_gaussian:
            if self.is_pure:
                A, B, C = bargmann.wigner_to_bargmann_psi(self.cov, self.means)
            else:
                A, B, C = bargmann.wigner_to_bargmann_rho(self.cov, self.means)
        else:
            return None
        if numpy:
            return math.asnumpy(A), math.asnumpy(B), math.asnumpy(C)
        return A, B, C

    def get_modes(self, item) -> State:
        r"""Returns the state on the given modes."""
        if isinstance(item, int):
            item = [item]
        elif isinstance(item, Iterable):
            item = list(item)
        else:
            raise TypeError("item must be int or iterable")

        if item == self.modes:
            return self

        if not set(item) & set(self.modes):
            raise ValueError(
                f"Failed to request modes {item} for state {self} on modes {self.modes}."
            )
        item_idx = [self.modes.index(m) for m in item]
        if self.is_gaussian:
            cov, _, _ = gaussian.partition_cov(self.cov, item_idx)
            means, _ = gaussian.partition_means(self.means, item_idx)
            return State(cov=cov, means=means, modes=item)

        fock_partitioned = fock.trace(self.dm(self.cutoffs), keep=item_idx)
        return State(dm=fock_partitioned, modes=item)

    # TODO: refactor
    def __eq__(self, other) -> bool:  # pylint: disable=too-many-return-statements
        r"""Returns whether the states are equal."""
        if self.num_modes != other.num_modes:
            return False
        if not np.isclose(self.purity, other.purity, atol=1e-6):
            return False
        if self.is_gaussian and other.is_gaussian:
            if not np.allclose(self.means, other.means, atol=1e-6):
                return False
            if not np.allclose(self.cov, other.cov, atol=1e-6):
                return False
            return True
        try:
            return np.allclose(
                self.ket(cutoffs=other.cutoffs),
                other.ket(cutoffs=other.cutoffs),
                atol=1e-6,
            )
        except TypeError:
            return np.allclose(
                self.dm(cutoffs=other.cutoffs),
                other.dm(cutoffs=other.cutoffs),
                atol=1e-6,
            )

    def __rshift__(self, other: Transformation) -> State:
        r"""Applies other (a Transformation) to self (a State), e.g., ``Coherent(x=0.1) >> Sgate(r=0.1)``."""
        if issubclass(other.__class__, State):
            raise TypeError(
                f"Cannot apply {other.__class__.__qualname__} to a state. Are you looking for the << operator?"
            )
        return other.primal(self)

    def __lshift__(self, other: State):
        r"""Implements projection onto a state or the dual transformation applied on a state.

        E.g., ``self << other`` where other is a ``State`` and ``self`` is either a ``State`` or a ``Transformation``.
        """
        return other.primal(self)

    def __add__(self, other: State):
        r"""Implements a mixture of states (only available in fock representation for the moment)."""
        if not isinstance(other, State):
            raise TypeError(f"Cannot add {other.__class__.__qualname__} to a state")
        warnings.warn("mixing states forces conversion to fock representation", UserWarning)
        return State(dm=self.dm(self.cutoffs) + other.dm(self.cutoffs))

    def __rmul__(self, other):
        r"""Implements multiplication by a scalar from the left.

        E.g., ``0.5 * psi``.
        """
        if self.is_gaussian:
            warnings.warn(
                "scalar multiplication forces conversion to fock representation",
                UserWarning,
            )
            if self.is_pure:
                return State(ket=self.ket() * other)
            return State(dm=self.dm() * other)
        if self._dm is not None:
            return State(dm=self.dm() * other, modes=self.modes)
        if self._ket is not None:
            return State(ket=self.ket() * other, modes=self.modes)
        raise ValueError("No fock representation available")

    def __truediv__(self, other):
        r"""Implements division by a scalar from the left.

        E.g. ``psi / 0.5``
        """
        if self.is_gaussian:
            warnings.warn("scalar division forces conversion to fock representation", UserWarning)
            if self.is_pure:
                return State(ket=self.ket() / other)
            return State(dm=self.dm() / other)
        if self._dm is not None:
            return State(dm=self.dm() / other, modes=self.modes)
        if self._ket is not None:
            return State(ket=self.ket() / other, modes=self.modes)
        raise ValueError("No fock representation available")

    @staticmethod
    def _format_probability(prob: float) -> str:
        if prob < 0.001:
            return f"{100*prob:.3e} %"
        else:
            return f"{prob:.3%}"

    def _repr_markdown_(self):
        table = (
            f"#### {self.__class__.__qualname__}\n\n"
            + "| Purity | Probability | Num modes | Bosonic size | Gaussian | Fock |\n"
            + "| :----: | :----: | :----: | :----: | :----: | :----: |\n"
            + f"| {self.purity :.2e} | "
            + self._format_probability(self.probability)
            + f" | {self.num_modes} | {'1' if self.is_gaussian else 'N/A'} | {'✅' if self.is_gaussian else '❌'} | {'✅' if self._ket is not None or self._dm is not None else '❌'} |"
        )

        if self.num_modes == 1:
            mikkel_plot(math.asnumpy(self.dm(cutoffs=self.cutoffs)))

        if settings.DEBUG:
            detailed_info = f"\ncov={repr(self.cov)}\n" + f"means={repr(self.means)}\n"
            return f"{table}\n{detailed_info}"

        return table


def mikkel_plot(
    rho: np.ndarray,
    xbounds: Tuple[int] = (-6, 6),
    ybounds: Tuple[int] = (-6, 6),
    **kwargs,
):  # pylint: disable=too-many-statements
    """Plots the Wigner function of a state given its density matrix.

    Args:
        rho (np.ndarray): density matrix of the state
        xbounds (Tuple[int]): range of the x axis
        ybounds (Tuple[int]): range of the y axis

    Keyword args:
        resolution (int): number of points used to calculate the wigner function
        xticks (Tuple[int]): ticks of the x axis
        xtick_labels (Optional[Tuple[str]]): labels of the x axis; if None uses default formatter
        yticks (Tuple[int]): ticks of the y axis
        ytick_labels (Optional[Tuple[str]]): labels of the y axis; if None uses default formatter
        grid (bool): whether to display the grid
        cmap (matplotlib.colormap): colormap of the figure

    Returns:
        tuple: figure and axes
    """

    plot_args = {
        "resolution": 200,
        "xticks": (-5, 0, 5),
        "xtick_labels": None,
        "yticks": (-5, 0, 5),
        "ytick_labels": None,
        "grid": False,
        "cmap": cm.RdBu,
    }
    plot_args.update(kwargs)

    if plot_args["xtick_labels"] is None:
        plot_args["xtick_labels"] = plot_args["xticks"]
    if plot_args["ytick_labels"] is None:
        plot_args["ytick_labels"] = plot_args["yticks"]

    q, ProbX = fock.quadrature_distribution(rho)
    p, ProbP = fock.quadrature_distribution(rho, np.pi / 2)

    xvec = np.linspace(*xbounds, plot_args["resolution"])
    pvec = np.linspace(*ybounds, plot_args["resolution"])
    W, X, P = wigner_discretized(rho, xvec, pvec)

    ### PLOTTING ###

    fig, ax = plt.subplots(
        2,
        2,
        figsize=(6, 6),
        gridspec_kw={"width_ratios": [2, 1], "height_ratios": [1, 2]},
    )
    plt.subplots_adjust(wspace=0.05, hspace=0.05)

    # Wigner function

    ax[1][0].contourf(X, P, W, 120, cmap=plot_args["cmap"], vmin=-abs(W).max(), vmax=abs(W).max())
    ax[1][0].set_xlabel("x", fontsize=12)
    ax[1][0].set_ylabel("p", fontsize=12)
    ax[1][0].get_xaxis().set_ticks(plot_args["xticks"])
    ax[1][0].xaxis.set_ticklabels(plot_args["xtick_labels"])
    ax[1][0].get_yaxis().set_ticks(plot_args["yticks"])
    ax[1][0].yaxis.set_ticklabels(plot_args["ytick_labels"], rotation="vertical", va="center")
    ax[1][0].tick_params(direction="in")
    ax[1][0].set_xlim(xbounds)
    ax[1][0].set_ylim(ybounds)
    ax[1][0].grid(plot_args["grid"])

    # X quadrature probability distribution
    ax[0][0].fill(q, ProbX, color=plot_args["cmap"](0.5))
    ax[0][0].plot(q, ProbX, color=plot_args["cmap"](0.8))
    ax[0][0].get_xaxis().set_ticks(plot_args["xticks"])
    ax[0][0].xaxis.set_ticklabels([])
    ax[0][0].get_yaxis().set_ticks([])
    ax[0][0].tick_params(direction="in")
    ax[0][0].set_ylabel("Prob(x)", fontsize=12)
    ax[0][0].set_xlim(xbounds)
    ax[0][0].set_ylim([0, 1.1 * max(ProbX)])
    ax[0][0].grid(plot_args["grid"])

    # P quadrature probability distribution
    ax[1][1].fill(ProbP, p, color=plot_args["cmap"](0.5))
    ax[1][1].plot(ProbP, p, color=plot_args["cmap"](0.8))
    ax[1][1].get_xaxis().set_ticks([])
    ax[1][1].get_yaxis().set_ticks(plot_args["yticks"])
    ax[1][1].yaxis.set_ticklabels([])
    ax[1][1].tick_params(direction="in")
    ax[1][1].set_xlabel("Prob(p)", fontsize=12)
    ax[1][1].set_xlim([0, 1.1 * max(ProbP)])
    ax[1][1].set_ylim(ybounds)
    ax[1][1].grid(plot_args["grid"])

    # Density matrix
    ax[0][1].matshow(abs(rho), cmap=plot_args["cmap"], vmin=-abs(rho).max(), vmax=abs(rho).max())
    ax[0][1].set_title("abs(ρ)", fontsize=12)
    ax[0][1].tick_params(direction="in")
    ax[0][1].get_xaxis().set_ticks([])
    ax[0][1].get_yaxis().set_ticks([])
    ax[0][1].set_aspect("auto")
    ax[0][1].set_ylabel(f"cutoff = {len(rho)}", fontsize=12)
    ax[0][1].yaxis.set_label_position("right")

    return fig, ax



# Copyright 2023 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""neighbours functions"""

from typing import Iterator

from numba import njit
from numba.cpython.unsafe.tuple import tuple_setitem

#################################################################################
## All neighbours means all the indices that differ from the given pivot by ±1 ##
#################################################################################


@njit
def all_neighbors(pivot: tuple[int, ...]) -> Iterator[tuple[int, tuple[int, ...]]]:
    r"""yields the indices of all the neighbours of the given index."""
    for j in range(len(pivot)):  # pylint: disable=consider-using-enumerate
        yield j, tuple_setitem(pivot, j, pivot[j] - 1)
        yield j, tuple_setitem(pivot, j, pivot[j] + 1)


####################################################################################
## Lower neighbours means all the indices that differ from the given index by -1  ##
####################################################################################


@njit
def lower_neighbors(pivot: tuple[int, ...]) -> Iterator[tuple[int, tuple[int, ...]]]:
    r"""yields the indices of the lower neighbours of the given index."""
    for j in range(len(pivot)):  # pylint: disable=consider-using-enumerate
        yield j, tuple_setitem(pivot, j, pivot[j] - 1)


####################################################################################
## Upper neighbours means all the indices that differ from the given index by +1  ##
####################################################################################


@njit
def upper_neighbors(pivot: tuple[int, ...]) -> Iterator[tuple[int, tuple[int, ...]]]:
    r"""yields the indices of the lower neighbours of the given index."""
    for j in range(len(pivot)):
        yield j, tuple_setitem(pivot, j, pivot[j] + 1)


####################################################################################################
## bitstring neighbours are indices that differ from the given index by ±1 according to a bitstring
####################################################################################################


@njit
def bitstring_neighbors(
    pivot: tuple[int, ...], bitstring: tuple[int, ...]
) -> Iterator[tuple[int, tuple[int, ...]]]:
    r"yields the indices of the bitstring neighbours of the given index"
    for i, b in enumerate(bitstring):
        if b:  # b == 1 -> lower
            yield i, tuple_setitem(pivot, i, pivot[i] - 1)
        else:  # b == 0 -> upper
            yield i, tuple_setitem(pivot, i, pivot[i] + 1)



# Copyright 2023 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=no-member

"""
This module contains the utility functions used by the classes in ``mrmustard.lab``.
"""
from typing import Callable, Optional, Tuple

from mrmustard.math.parameters import update_euclidean, Constant, Variable


def make_parameter(
    is_trainable: bool,
    value: any,
    name: str,
    bounds: Tuple[Optional[float], Optional[float]],
    update_fn: Callable = update_euclidean,
):
    r"""
    Returns a constant or variable parameter with given name, value, bounds, and update function.

    Args:
        is_trainable: Whether to return a variable (``True``) or constant (``False``) parameter.
        value: The value of the returned parameter.
        name: The name of the returned parameter.
        bounds: The bounds of the returned parameter (ignored if ``is_trainable`` is ``False``).
        update_fn: The update_fn of the returned parameter (ignored if ``is_trainable`` is ``False``).
    """
    if isinstance(value, (Constant, Variable)):
        return value
    if not is_trainable:
        return Constant(value=value, name=name)
    return Variable(value=value, name=name, bounds=bounds, update_fn=update_fn)



# Copyright 2021 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This module implements the :class:`.Circuit` class which acts as a representation for quantum circuits.
"""

from __future__ import annotations

__all__ = ["Circuit"]

from typing import List, Optional, Tuple

from mrmustard import settings
from mrmustard.lab.abstract import State, Transformation
from mrmustard.utils.typing import RealMatrix, RealVector
from mrmustard.lab.circuit_drawer import circuit_text
from mrmustard.math.tensor_wrappers import XPMatrix, XPVector

import numpy as np


class Circuit(Transformation):
    """Represents a quantum circuit: a set of operations to be applied on quantum states.

    Args:
        ops (list or none): A list of operations comprising the circuit.
    """

    def __init__(self, ops: Optional[List] = None):
        self._ops = list(ops) if ops is not None else []
        super().__init__(name="Circuit")
        self.reset()

    @property
    def ops(self) -> Optional[List]:
        r"""
        The list of operations comprising the circuit.
        """
        return self._ops

    def reset(self):
        """Resets the state of the circuit clearing the list of modes and setting the compiled flag to false."""
        self._compiled: bool = False
        self._modes: List[int] = []

    @property
    def num_modes(self) -> int:
        all_modes = {mode for op in self._ops for mode in op.modes}
        return len(all_modes)

    def primal(self, state: State) -> State:
        for op in self._ops:
            state = op.primal(state)
        return state

    def dual(self, state: State) -> State:
        for op in reversed(self._ops):
            state = op.dual(state)
        return state

    def XYd(
        self,
        allow_none: bool = True,
    ) -> Tuple[
        RealMatrix, RealMatrix, RealVector
    ]:  # NOTE: Overriding Transformation.XYd for efficiency
        X = XPMatrix(like_1=True)
        Y = XPMatrix(like_0=True)
        d = XPVector()
        for op in self._ops:
            opx, opy, opd = op.XYd(allow_none)
            opX = XPMatrix.from_xxpp(opx, modes=(op.modes, op.modes), like_1=True)
            opY = XPMatrix.from_xxpp(opy, modes=(op.modes, op.modes), like_0=True)
            opd = XPVector.from_xxpp(opd, modes=op.modes)
            if opX.shape is not None and opX.shape[-1] == 1 and len(op.modes) > 1:
                opX = opX.clone(len(op.modes), modes=(op.modes, op.modes))
            if opY.shape is not None and opY.shape[-1] == 1 and len(op.modes) > 1:
                opY = opY.clone(len(op.modes), modes=(op.modes, op.modes))
            if opd.shape is not None and opd.shape[-1] == 1 and len(op.modes) > 1:
                opd = opd.clone(len(op.modes), modes=op.modes)
            X = opX @ X
            Y = opX @ Y @ opX.T + opY
            d = opX @ d + opd
        return X.to_xxpp(), Y.to_xxpp(), d.to_xxpp()

    @property
    def is_gaussian(self):
        """Returns `true` if all operations in the circuit are Gaussian."""
        return all(op.is_gaussian for op in self._ops)

    @property
    def is_unitary(self):
        """Returns `true` if all operations in the circuit are unitary."""
        return all(op.is_unitary for op in self._ops)

    def value(self, shape: Tuple[int]):
        raise NotImplementedError

    def __len__(self):
        return len(self._ops)

    _repr_markdown_ = None

    def __repr__(self) -> str:
        """String to display the object on the command line."""
        # ops_repr = [repr(op) for op in self._ops]
        # return " >> ".join(ops_repr)
        return circuit_text(self._ops, decimals=settings.CIRCUIT_DECIMALS)

    def __str__(self):
        """String representation of the circuit."""
        ops_repr = [repr(op) for op in self._ops]
        return " >> ".join(ops_repr)

    # pylint: disable=too-many-branches,too-many-return-statements
    def __eq__(self, other):
        r"""Returns ``True`` if the two transformations are equal."""
        if not isinstance(other, Circuit):
            return False
        if not (self.is_gaussian and other.is_gaussian):
            return np.allclose(
                self.choi(cutoffs=[settings.EQ_TRANSFORMATION_CUTOFF] * 4 * self.num_modes),
                other.choi(cutoffs=[settings.EQ_TRANSFORMATION_CUTOFF] * 4 * self.num_modes),
                rtol=settings.EQ_TRANSFORMATION_RTOL_FOCK,
            )

        sX, sY, sd = self.XYd(allow_none=False)
        oX, oY, od = other.XYd(allow_none=False)
        return np.allclose(sX, oX) and np.allclose(sY, oY) and np.allclose(sd, od)



# Copyright 2023 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

" This module contains binomial strategies "

from typing import Optional

import numpy as np
from numba import njit, typed, types

from mrmustard.math.lattice import paths, steps
from mrmustard.utils.typing import ComplexMatrix, ComplexTensor, ComplexVector

SQRT = np.sqrt(np.arange(100000))

__all__ = ["binomial", "binomial_dict", "binomial_numba"]


def binomial(
    local_cutoffs: tuple[int, ...],
    A: ComplexMatrix,
    b: ComplexVector,
    c: complex,
    max_l2: float,
    global_cutoff: int,
) -> ComplexTensor:
    r"""Binomial strategy (fill ket by weight), python version with numba function/loop.

    Args:
        local_cutoffs (tuple[int, ...]): local cutoffs of the tensor (used at least as shape)
        A (np.ndarray): A matrix of the Fock-Bargmann representation
        b (np.ndarray): B vector of the Fock-Bargmann representation
        c (complex): vacuum amplitude
        max_l2 (float): max L2 norm. If reached, the computation is stopped early.
        global_cutoff (Optional[int]): global cutoff (max total photon number considered + 1).

    Returns:
        G, prob (np.ndarray, float): Fock representation of the Gaussian tensor with shape ``shape`` and L2 norm
    """
    # init output tensor
    G = np.zeros(local_cutoffs, dtype=np.complex128)

    # write vacuum amplitude
    G.flat[0] = c
    norm = np.abs(c) ** 2

    # iterate over subspaces by weight and stop if norm is large enough. Caches indices.
    for photons in range(1, global_cutoff):
        try:
            indices = paths.BINOMIAL_PATHS_PYTHON[(local_cutoffs, photons)]
        except KeyError:
            indices = paths.binomial_subspace_basis(local_cutoffs, photons)
            paths.BINOMIAL_PATHS_PYTHON[(local_cutoffs, photons)] = indices
        G, subspace_norm = steps.binomial_step(G, A, b, indices)  # numba parallelized function
        norm += subspace_norm
        try:
            if norm > max_l2:
                break
        except TypeError:  # max_l2 is None
            pass
    return G, norm


def binomial_dict(
    local_cutoffs: tuple[int, ...],
    A: ComplexMatrix,
    b: ComplexVector,
    c: complex,
    max_prob: Optional[float] = None,
    global_cutoff: Optional[int] = None,
) -> dict[tuple[int, ...], complex]:
    r"""Factorial speedup strategy (fill ket by weight), python version with numba function/loop.
    Uses a dictionary to store the output.

    Args:
        local_cutoffs (tuple[int, ...]): local cutoffs of the tensor (used at least as shape)
        A (np.ndarray): A matrix of the Fock-Bargmann representation
        b (np.ndarray): B vector of the Fock-Bargmann representation
        c (complex): vacuum amplitude
        max_prob (float): max L2 norm. If reached, the computation is stopped early.
        global_cutoff (Optional[int]): global cutoff (max total photon number considered).
            If not given it is calculated from the local cutoffs.

    Returns:
        dict[tuple[int, ...], complex]: Fock representation of the Gaussian tensor.
    """
    if global_cutoff is None:
        global_cutoff = sum(local_cutoffs) - len(local_cutoffs)

    # init numba output dict
    G = typed.Dict.empty(
        key_type=types.UniTuple(types.int64, len(local_cutoffs)),
        value_type=types.complex128,
    )

    # write vacuum amplitude
    G[(0,) * len(local_cutoffs)] = c
    prob = np.abs(c) ** 2

    # iterate over subspaces by weight and stop if norm is large enough. Caches indices.
    for photons in range(1, global_cutoff):
        try:
            indices = paths.BINOMIAL_PATHS_PYTHON[(local_cutoffs, photons)]
        except KeyError:
            indices = paths.binomial_subspace_basis(local_cutoffs, photons)
            paths.BINOMIAL_PATHS_PYTHON[(local_cutoffs, photons)] = indices
        G, prob_subspace = steps.binomial_step_dict(G, A, b, indices)  # numba parallelized function
        prob += prob_subspace
        try:
            if prob > max_prob:
                break
        except TypeError:
            pass
    return G


@njit
def binomial_numba(
    local_cutoffs: tuple[int, ...],
    A: ComplexMatrix,
    b: ComplexVector,
    c: complex,
    FP: dict[tuple[tuple[int, ...], int], list[tuple[int, ...]]],
    max_prob: float = 0.999,
    global_cutoff: Optional[int] = None,
) -> ComplexTensor:  # pragma: no cover
    r"""Binomial strategy (fill by weight), fully numba version."""
    if global_cutoff is None:
        global_cutoff = sum(local_cutoffs) - len(local_cutoffs)

    # init output tensor
    G = np.zeros(local_cutoffs, dtype=np.complex128)

    # write vacuum amplitude
    G.flat[0] = c
    prob = np.abs(c) ** 2

    # iterate over all other indices in parallel and stop if norm is large enough
    for photons in range(1, global_cutoff):
        try:
            indices = FP[(local_cutoffs, photons)]
        except Exception:  # pylint: disable=broad-except
            indices = paths.binomial_subspace_basis(local_cutoffs, photons)
            FP[(local_cutoffs, photons)] = indices
        G, prob_subspace = steps.binomial_step(G, A, b, indices)
        prob += prob_subspace
        if prob > max_prob:
            break
    return G



# Copyright 2021 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

r"""
The point of entry for the backend.
"""
import sys

from .autocast import *
from .caching import *
from .backend_base import *
from .backend_manager import BackendManager
from .backend_numpy import *
from .lattice import *
from .parameters import *
from .parameter_set import *
from .tensor_networks import *
from .tensor_wrappers import *

sys.modules[__name__] = BackendManager()



# Copyright 2023 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

" This module contains strategies for calculating the matrix elements of the squeezing gate. "

import numpy as np
from numba import njit

from mrmustard.math.lattice import steps
from mrmustard.utils.typing import ComplexTensor

SQRT = np.sqrt(np.arange(100000))

__all__ = ["squeezer", "squeezer_vjp", "squeezed", "squeezed_vjp"]


@njit
def squeezer(
    shape: tuple[int, int], r: float, theta: float, dtype=np.complex128
):  # pragma: no cover
    r"""Calculates the matrix elements of the squeezing gate using a recurrence relation.
    (See eq. 50-52 in https://arxiv.org/abs/2004.11002)
        Args:
            shape (tuple[int, int]): Fock cutoffs for the output and input indices.
            r (float): squeezing magnitude
            theta (float): squeezing angle
            dtype (data type): data type used for the calculation.

        Returns:
            array (ComplexMatrix): matrix representing the squeezing gate.
    """
    M, N = shape
    S = np.zeros(shape, dtype=dtype)

    eitheta_tanhr = np.exp(1j * theta) * np.tanh(r)
    eitheta_tanhr_conj = np.conj(eitheta_tanhr)
    sechr = 1.0 / np.cosh(r)

    S[0, 0] = np.sqrt(sechr)
    for m in range(2, M, 2):
        S[m, 0] = -SQRT[m - 1] / SQRT[m] * eitheta_tanhr * S[m - 2, 0]

    for m in range(M):
        for n in range(2 - (m % 2), N, 2):
            # for n in range(1, N):
            if (m + n) % 2 == 0:
                S[m, n] = (
                    SQRT[n - 1] / SQRT[n] * eitheta_tanhr_conj * S[m, n - 2]
                    + SQRT[m] / SQRT[n] * sechr * S[m - 1, n - 1]
                )
    return S


@njit
def squeezer_vjp(
    G: ComplexTensor,
    dLdG: ComplexTensor,
    r: float,
    phi: float,
) -> tuple[float, float]:  # pragma: no cover
    r"""Squeezing gradients with respect to r and theta.
    This function could return dL/dA, dL/db, dL/dc like its vanilla counterpart,
    but it is more efficient to include this chain rule step in the numba function, since we can.

    Args:
        G (np.ndarray): Tensor result of the forward pass
        dLdG (np.ndarray): gradient of the loss with respect to the output tensor
        r (float): squeezing magnitude
        phi (float): squeezing angle

    Returns:
        tuple[float, float]: dL/dr, dL/phi
    """
    M, N = G.shape

    # init gradients
    dA = np.zeros((2, 2), dtype=np.complex128)  # dGdA at an index (of G)
    _ = np.zeros(2, dtype=np.complex128)
    dLdA = np.zeros_like(dA)

    # first column
    for m in range(2, M, 2):
        dA, _ = steps.vanilla_step_grad(G, (m, 0), dA, _)
        dLdA += dA * dLdG[m, 0]

    # rest of the matrix
    for m in range(M):
        for n in range(1, N):
            if (m + n) % 2 == 0:
                dA, _ = steps.vanilla_step_grad(G, (m, n), dA, _)
                dLdA += dA * dLdG[m, n]

    dLdC = np.sum(G * dLdG)  # np.sqrt(np.cosh(r)) cancels out with 1 / np.sqrt(np.cosh(r)) later
    # chain rule
    d_sech = -np.tanh(r) / np.cosh(r)
    d_tanh = 1.0 / np.cosh(r) ** 2
    tanh = np.tanh(r)
    exp = np.exp(1j * phi)
    exp_conj = np.exp(-1j * phi)

    dLdr = 2 * np.real(
        -dLdA[0, 0] * exp * d_tanh
        + dLdA[0, 1] * d_sech
        + dLdA[1, 1] * exp_conj * d_tanh
        - np.conj(dLdC) * 0.5 * tanh  # / np.sqrt(np.cosh(r))
    )
    dLdphi = 2 * np.real(-dLdA[0, 0] * 1j * exp * tanh - dLdA[1, 1] * 1j * exp_conj * tanh)

    return dLdr, dLdphi


@njit
def squeezed(cutoff: int, r: float, theta: float, dtype=np.complex128):  # pragma: no cover
    r"""Calculates the matrix elements of the single-mode squeezed state using recurrence relations.

    Args:
        cutoff (int): Fock cutoff for the ket
        r (float): squeezing magnitude
        theta (float): squeezing angle
        dtype (data type): data type used for the calculation.

    Returns:
        array (ComplexMatrix): matrix representing the squeezing gate.
    """
    S = np.zeros(cutoff, dtype=dtype)
    eitheta_tanhr = np.exp(1j * theta) * np.tanh(r)
    S[0] = np.sqrt(1.0 / np.cosh(r))

    for m in range(2, cutoff, 2):
        S[m] = SQRT[m - 1] / SQRT[m] * eitheta_tanhr * S[m - 2]

    return S


@njit
def squeezed_vjp(
    G: ComplexTensor,
    dLdG: ComplexTensor,
    r: float,
    phi: float,
) -> tuple[float, float]:  # pragma: no cover
    r"""Squeezed state gradients with respect to r and theta.
    This function could return dL/dA, dL/db, dL/dc like its vanilla counterpart,
    but it is more efficient to include this chain rule step in the numba function, since we can.

    Args:
        G (np.ndarray): Tensor result of the forward pass
        dLdG (np.ndarray): gradient of the loss with respect to the output tensor
        r (float): squeezing magnitude
        phi (float): squeezing angle

    Returns:
        tuple[float, float]: dL/dr, dL/phi
    """
    M = G.shape[0]

    # init gradients
    dA = np.zeros((1, 1), dtype=np.complex128)
    _ = np.zeros(1, dtype=np.complex128)
    dLdA = np.zeros_like(dA)

    # first column
    for m in range(2, M, 2):
        dA, _ = steps.vanilla_step_grad(G, (m,), dA, _)
        dLdA += dA * dLdG[m]

    # chain rule
    tanh = np.tanh(r)
    d_tanh = 1.0 / np.cosh(r) ** 2
    exp = np.exp(1j * phi)

    dLdC = np.sum(G * dLdG)  # np.sqrt(np.cosh(r)) cancels out with 1 / np.sqrt(np.cosh(r)) later

    dLdr = 2 * np.real(-dLdA[0, 0] * exp * d_tanh - np.conj(dLdC) * 0.5 * tanh)
    dLdphi = 2 * np.real(-dLdA[0, 0] * 1j * exp * tanh)

    return dLdr, dLdphi
